
DEFAULT_PROFILE = 'ConfigFileDefaultProfile'


class ConfigFileDefaultProfile(object):
    PROFILE_FROM = 'config-file-default'


class AppInitProfile(object):
    PROFILE_FROM = 'app-init'


class TestProfile(object):
    PROFILE_FROM = 'testing-default'


class EnvironmentProfile(object):
    PROFILE_FROM = 'environment'


class CLIOption(object):
    PROFILE_FROM = 'cli-option'



import inspect
import os
import weakref

import flask
import pathlib
from werkzeug.utils import import_string

from keg.extensions import lazy_gettext as _

_signature_cache = weakref.WeakKeyDictionary()


# sentinal object
class NotGiven(object):
    pass


def ensure_dirs(newdir, mode=NotGiven):
    """
        A "safe" verision of Path.makedir(..., parents=True) that will only create the directory
        if it doesn't already exist.  We also manually create parents so that mode is set
        correctly.  Python docs say that mode is ignored when using Path.mkdir(..., parents=True)
    """
    if mode is NotGiven:
        mode = flask.current_app.config['KEG_DIR_MODE']
    dpath = pathlib.Path(newdir)
    if dpath.is_dir():
        return
    if dpath.is_file():
        raise OSError(_("A file with the same name as the desired"
                        " directory, '{dpath}', already exists.", dpath=dpath))

    ensure_dirs(dpath.parent, mode=mode)
    dpath.mkdir(mode)


class ClassProperty(property):
    """A decorator that behaves like @property except that operates
    on classes rather than instances.
    """
    def __init__(self, fget, *arg, **kw):
        self.ignore_set = kw.pop('ignore_set', False)
        self.__doc__ = fget.__doc__
        super(ClassProperty, self).__init__(fget, *arg, **kw)

    def __get__(self, obj, cls):  # noqa
        return self.fget(cls)

    def __set__(self, cls, value):  # noqa
        if self.fset is None and not self.ignore_set:
            raise AttributeError(_("can't set attribute"))
        if not self.ignore_set:
            self.fset(cls, value)


classproperty = ClassProperty


class HybridMethod(object):
    """
    A decorator which allows definition of a Python object method with both
    instance-level and class-level behavior::

        Class Bar:
            @hybridmethod
            def foo(self, rule, **options):
                # this is used in an instance context

            @foo.classmethod
            def foo(cls, rule, **options):
                # this is used in class context
    """

    def __init__(self, func, cm_func=None):
        self.instance_func = func
        self.classmethod(cm_func or func)

    def __get__(self, instance, owner):
        if instance is None:
            return self.cm_func.__get__(owner, owner.__class__)
        else:
            return self.instance_func.__get__(instance, owner)

    def classmethod(self, cm_func):
        """Provide a modifying decorator that is used as a classmethod decorator."""

        self.cm_func = cm_func
        if not self.cm_func.__doc__:
            self.cm_func.__doc__ = self.instance_func.__doc__
        return self


hybridmethod = HybridMethod


def pymodule_fpaths_to_objects(fpaths):
    """
        Takes an iterable of file paths reprenting possible python modules and will return an
        iterable of tuples with the file path along with the contents of that file if the file
        exists.

        If the file does not exist or cannot be accessed, the third term of the tuple stores
        the exception.
    """
    retval = []
    for fpath in fpaths:
        try:
            pymodule_globals = {}
            with open(fpath) as fo:
                exec(compile(fo.read(), fpath, 'exec'), pymodule_globals)
            retval.append((fpath, pymodule_globals, None))
        except (FileNotFoundError, IsADirectoryError, PermissionError) as exc:
            retval.append((fpath, None, exc))
    return retval


def app_environ_get(app_import_name, key, default=None):
    # App names often have periods and it is not possible to export an
    # environment variable with a period in it.
    app_name = app_import_name.replace('.', '_').upper()
    environ_key = '{}_{}'.format(app_name, key.upper())

    return os.environ.get(environ_key, default)


def visit_modules(dotted_paths, base_path=None):
    for path in dotted_paths:
        if path.startswith('.') and base_path is not None:
            path = base_path + path
        import_string(path)


def validate_arguments(func, args, kwargs, drop_extra=True):  # type: ignore
    """Checks if the function accepts the arguments and keyword arguments.
    Returns a new ``(args, kwargs)`` tuple that can safely be passed to
    the function without causing a `TypeError` because the function signature
    is incompatible.  If `drop_extra` is set to `True` (which is the default)
    any extra positional or keyword arguments are dropped automatically.

    The exception raised provides three attributes:

    `missing`
        A set of argument names that the function expected but where
        missing.

    `extra`
        A dict of keyword arguments that the function can not handle but
        where provided.

    `extra_positional`
        A list of values that where given by positional argument but the
        function cannot accept.

    This can be useful for decorators that forward user submitted data to
    a view function::

        from werkzeug.utils import ArgumentValidationError, validate_arguments

        def sanitize(f):
            def proxy(request):
                data = request.values.to_dict()
                try:
                    args, kwargs = validate_arguments(f, (request,), data)
                except ArgumentValidationError:
                    raise BadRequest('The browser failed to transmit all '
                                     'the data expected.')
                return f(*args, **kwargs)
            return proxy

    :param func: the function the validation is performed against.
    :param args: a tuple of positional arguments.
    :param kwargs: a dict of keyword arguments.
    :param drop_extra: set to `False` if you don't want extra arguments
                       to be silently dropped.
    :return: tuple in the form ``(args, kwargs)``.

    .. deprecated:: 2.0
        Will be removed in Werkzeug 2.1. Use :func:`inspect.signature`
        instead.

    Copied from Werkzeug
    """
    parser = _parse_signature(func)
    args, kwargs, missing, extra, extra_positional = parser(args, kwargs)[:5]
    if missing:
        raise ArgumentValidationError(tuple(missing))
    elif (extra or extra_positional) and not drop_extra:
        raise ArgumentValidationError(None, extra, extra_positional)
    return tuple(args), kwargs


def _parse_signature(func):  # type: ignore
    """Return a signature object for the function.

    .. deprecated:: 2.0
        Will be removed in Werkzeug 2.1 along with ``utils.bind`` and
        ``validate_arguments``.
    """
    # if we have a cached validator for this function, return it
    parse = _signature_cache.get(func)
    if parse is not None:
        return parse

    # inspect the function signature and collect all the information
    tup = inspect.getfullargspec(func)
    positional, vararg_var, kwarg_var, defaults = tup[:4]
    defaults = defaults or ()
    arg_count = len(positional)
    arguments = []
    for idx, name in enumerate(positional):
        if isinstance(name, list):
            raise TypeError(
                "cannot parse functions that unpack tuples in the function signature"
            )
        try:
            default = defaults[idx - arg_count]
        except IndexError:
            param = (name, False, None)
        else:
            param = (name, True, default)
        arguments.append(param)
    arguments = tuple(arguments)

    def parse(args, kwargs):  # type: ignore
        new_args = []
        missing = []
        extra = {}

        # consume as many arguments as positional as possible
        for idx, (name, has_default, default) in enumerate(arguments):
            try:
                new_args.append(args[idx])
            except IndexError:
                try:
                    new_args.append(kwargs.pop(name))
                except KeyError:
                    if has_default:
                        new_args.append(default)
                    else:
                        missing.append(name)
            else:
                if name in kwargs:
                    extra[name] = kwargs.pop(name)

        # handle extra arguments
        extra_positional = args[arg_count:]
        if vararg_var is not None:
            new_args.extend(extra_positional)
            extra_positional = ()
        if kwargs and kwarg_var is None:
            extra.update(kwargs)
            kwargs = {}

        return (
            new_args,
            kwargs,
            missing,
            extra,
            extra_positional,
            arguments,
            vararg_var,
            kwarg_var,
        )

    _signature_cache[func] = parse
    return parse


class ArgumentValidationError(ValueError):
    """Raised if :func:`validate_arguments` fails to validate

    .. deprecated:: 2.0
        Will be removed in Werkzeug 2.1 along with ``utils.bind`` and
        ``validate_arguments``.
    """

    def __init__(self, missing=None, extra=None, extra_positional=None):  # type: ignore
        self.missing = set(missing or ())
        self.extra = extra or {}
        self.extra_positional = extra_positional or []
        super().__init__(
            "function arguments invalid."
            f" ({len(self.missing)} missing,"
            f" {len(self.extra) + len(self.extra_positional)} additional)"
        )



from __future__ import absolute_import

import flask

from keg.web import BaseView as KegBaseView

blueprint = flask.Blueprint(
    'custom',
    __name__,
    template_folder='../templates/specific-path',
    url_prefix='/tanagra',
)


class BaseView(KegBaseView):
    blueprint = blueprint


class BlueprintTest(BaseView):
    def get(self):
        return self.render()


class BlueprintTest2(BaseView):
    url = '/custom-route'

    def get(self):
        return self.render()



from unittest import mock

import pytest
import sqlalchemy as sa

from keg.app import Keg, KegAppError
from keg_apps.cli2.app import CLI2App
from keg_apps.db2 import DB2App


class TestInit(object):

    def test_init_called_twice_error(self):
        with pytest.raises(KegAppError, match=r'init\(\) already called on this instance'):
            app = CLI2App(__name__)
            app.init(use_test_profile=True)
            app.init(use_test_profile=True)


class TestSQLitePragma:
    @mock.patch('keg.db.sa_event', autospec=True, spec_set=True)
    def test_config_set(self, m_sa_event):
        DB2App.testing_prep(KEG_SQLITE_ENABLE_FOREIGN_KEYS=True)
        m_sa_event.listens_for.assert_called_once_with(sa.engine.Engine, 'connect')

    @mock.patch('keg.db.sa_event', autospec=True, spec_set=True)
    def test_config_not_set(self, m_sa_event):
        DB2App.testing_prep()
        m_sa_event.listens_for.assert_not_called()


class TestRouteDecorator(object):
    class TRDApp(Keg):
        import_name = __name__ + ':TRDApp'

    class TRDApp2(Keg):
        import_name = __name__ + ':TRDApp2'

    def test_class_method(self):
        @self.TRDApp.route('/foo')
        def foo():
            pass

        @self.TRDApp.route('/bar')
        def bar():
            pass

        @self.TRDApp2.route('/baz')
        def baz():
            pass

        assert len(self.TRDApp._routes) == 2
        assert len(self.TRDApp2._routes) == 1



from keg.web import BaseView

from . import __component__

component_bp = __component__.create_named_blueprint(__name__)


class Blog(BaseView):
    blueprint = component_bp

    def get(self):
        return self.render()



from __future__ import absolute_import
import flask

from flask import request
from keg.web import BaseView as KegBaseView, route, rule

from keg_apps.extensions import gettext as _


blueprint = flask.Blueprint('routing', __name__)


class BaseView(KegBaseView):
    blueprint = blueprint


class VerbRouting(BaseView):
    """
        GET /verb-routing -> 'method get'
        POST /verb-routing -> 'method post'
    """

    def get(self):
        return _('method get')

    def post(self):
        return _('method post')


class VerbRoutingSub(VerbRouting):
    """
        Same as VerbRouting
    """


class ExplicitRoute(BaseView):
    """
        /explicit-route -> 404
        /some-route/foo -> 'get some-route'
    """
    # absolute URL indicates the default class URL will not be generated
    rule('/some-route')

    def get(self):
        return _('get some-route')


class ExplicitRouteAlt(KegBaseView):
    """
        /some-route-alt -> 'get some-route alt'

        Same as ExplicitRoute, but blueprint is not assigned in class definition. For route
        to be available, assign_blueprint must be called prior to blueprint registration.
    """
    rule('/some-route-alt')

    def get(self):
        return _('get some-route alt')


class HelloWorld(BaseView):
    """
        /hello-world -> 'Hello World'
        /hello-world/foo -> 'Hello Foo'
    """
    rule()
    rule('<name>')

    def get(self, name='World'):
        return _('Hello {name}', name=name)


class HWRuleDefault(BaseView):
    """
        GET /hwrd -> 'Hello RD'
        GET /hwrd/foo -> 'Hello Foo'
        POST /hwrd -> 'post Hello RD'
        POST /hwrd-> 'post Hello Foo'
    """
    # Since the default rule needs to have defaults, you need to define it and not rely on the
    # auto-generated one.
    rule('/hwrd', post=True, defaults={'name': 'RD'})
    rule('<name>', post=True)

    def post(self, name):
        return _('post Hello {name}', name=name)

    def get(self, name):
        return _('Hello {name}', name=name)


class HelloReq(BaseView):
    """
        /hello-req -> 404
        /hello-req/foo -> 'Hello Foo'
    """
    # making the rule absolute disables the default rule that would have been created by the class.
    rule('/hello-req/<name>')

    def get(self, name):
        return _('Hello {name}', name=name)


class HelloReq2(BaseView):
    """
        /hello-req2 -> 404
        /hello-req2/foo -> 'Hello Foo'
    """
    # no absolute rule, but only one endpoint to use
    rule('<name>')

    def get(self, name):
        return _('Hello {name}', name=name)


class Cars(BaseView):
    """
        CRUD for a model/entity

        GET /cars/list - show list of cars that can be managed
        GET /cars/create - show empty car form
        POST /cars/create - create a new car
        GET /cars/edit/12 - show car form w/ data
        POST /cars/edit/12 - update car 12
        GET /cars/delete/12 - deletes car 12
    """

    @route()
    def list(self):
        return _('list')

    @route()
    def create_get(self):
        return _('create get')

    @route(post_only=True)
    def create_post(self):
        return _('create post')

    @route('<int:car_id>', post=True)
    def edit(self, car_id):
        if request.method == 'GET':
            return _('form w/ data: {id}', id=car_id)
        else:
            return _('update car: {id}', id=car_id)

    @route('<int:ident>')
    def delete(self, ident):
        return _('delete: {ident}', ident=ident)


class Tickets(BaseView):
    """
        REST API Example

        GET /tickets - Retrieves a list of tickets
        GET /tickets/12 - Retrieves a specific ticket
        POST /tickets - Creates a new ticket
        PUT /tickets/12 - Updates ticket #12
        PATCH /tickets/12 - Partially updates ticket #12
        DELETE /tickets/12 - Deletes ticket #12
    """
    rule('/tickets', methods=['GET', 'POST'])
    rule('<int:ticket_id>', methods=['GET', 'PUT', 'PATCH', 'DELETE'])

    def get(self, ticket_id=None):
        if ticket_id:
            return _('single')
        return _('list')

    def post(self):
        return _('create')

    def put(self, ticket_id):
        return _('update')

    def patch(self, ticket_id):
        return _('partial update')

    def delete(self, ticket_id):
        return _('delete')


class Misc(BaseView):
    rule('/misc')
    rule('foo')
    rule('/misc2')
    rule('bar', post_only=True)

    def get(self):
        return _('get')

    def post(self):
        return _('post')

    @route('/an-abs-url')
    def an_abs_url(self):
        return _('found me')

    @route('8')
    @route('9')
    def two_routes(self):
        return '17'


class CrudBase(KegBaseView):
    """
        Testing a view that is intended to be used as an abstract class: it should be inherited
        but will never be instantiated itself.

        This is similiar to BaseView, except that this class is intended to represent a class that
        isn't confined to a certain blueprint, but intended to give similar functionality in a way
        that can be used in/with different blueprints.

        Pain points that this identifies:

            - This class itself should be created without throwing an exception due to not having
              a blueprint.
            - The creation of this class should not cause any application routes to be defined.
            - The creation of a subclass of this class should result in routes for the subclass
              being created.
    """
    @route()
    def list(self):
        return _('listing {class_name}', class_name=self.__class__.__name__)


class Trucks(CrudBase):
    """
        CRUD for a model/entity

        GET /trucks/list
    """
    blueprint = blueprint


class Planes(CrudBase):
    """
        CRUD for a model/entity

        GET /planes/list
    """
    blueprint = blueprint



import pytest

from keg import current_app
from keg.db import db
from keg.signals import db_init_pre, db_init_post, db_clear_post, db_clear_pre
from keg.testing import invoke_command

# Note: we do not want to import Blog directly here. Doing so will affect metadata collection and
# prevent the tests below from proving that component initialization is loading the model.
# from keg_apps.db.blog.model.entities import Blog
import keg_apps.db.model.entities as ents
from keg_apps.db.app import DBApp
from keg_apps.db2 import DB2App


@pytest.fixture(autouse=True)
def db_session_prep():
    """
        Rollback the session after every test.
    """
    db.session.rollback()


class TestDB(object):

    @classmethod
    def setup_class(cls):
        DBApp.testing_prep()

    def test_primary_db_entity(self):
        # importing Blog here to ensure we're importing after app init, which will only succeed if
        # the component has loaded properly
        from keg_apps.db.blog.model.entities import Blog
        assert Blog.query.count() == 0
        blog = Blog(title='foo')
        db.session.add(blog)
        db.session.commit()
        assert Blog.query.count() == 1

    def test_postgres_bind_db_entity(self):
        assert ents.PGDud.query.count() == 0
        dud = ents.PGDud(name='foo')
        db.session.add(dud)
        db.session.commit()
        assert ents.PGDud.query.count() == 1

    def test_postgres_db_enttity_alt_schema(self):
        assert ents.PGDud2.query.count() == 0
        dud = ents.PGDud2(name='foo')
        db.session.add(dud)
        db.session.commit()
        assert ents.PGDud2.query.count() == 1


class TestDB2(object):

    def test_init_without_db_binds(self):
        # Make sure we don't get an error initializing the app when the SQLALCHEMY_BINDS config
        # option is None
        app = DB2App.testing_prep()
        value = app.config.get('SQLALCHEMY_BINDS')
        # flask-sqlalchemy < 3.0: None
        # flask-sqlalchemy 3.0+: {}
        assert value is None or value == {}


class TestDatabaseManager(object):
    @classmethod
    def setup_class(cls):
        DBApp.testing_prep()

    def test_db_init_with_clear(self):
        # importing Blog here to ensure we're importing after app init, which will only succeed if
        # the component has loaded properly
        from keg_apps.db.blog.model.entities import Blog

        # have to use self here to enable the inner functions to adjust an outer-context variable
        self.init_pre_connected = False
        self.init_post_connected = False
        self.clear_pre_connected = False
        self.clear_post_connected = False

        @db_init_pre.connect_via(current_app._get_current_object())
        def catch_init_pre(app):
            self.init_pre_connected = True

        @db_init_post.connect_via(current_app._get_current_object())
        def catch_init_post(app):
            self.init_post_connected = True

        @db_clear_pre.connect_via(current_app._get_current_object())
        def catch_clear_pre(app):
            self.clear_pre_connected = True

        @db_clear_post.connect_via(current_app._get_current_object())
        def catch_clear_post(app):
            self.clear_post_connected = True

        Blog.testing_create()
        assert Blog.query.count() >= 1
        ents.PGDud2.testing_create()
        assert ents.PGDud2.query.count() >= 1

        current_app.db_manager.db_init_with_clear()

        # todo: We could check all the intermediate steps, but this is more a functional test
        # for now.  If the record is gone and we can create an new blog post, then we assume the
        # clear and init went ok.
        assert Blog.query.count() == 0
        Blog.testing_create()
        assert Blog.query.count() == 1

        assert self.init_pre_connected
        assert self.init_post_connected
        assert self.clear_pre_connected
        assert self.clear_post_connected


class TestKegSQLAlchemy(object):
    def test_session_ids_with_cli_invocation(self):
        sess_id = id(db.session)

        result = invoke_command(DBApp, 'hello')
        assert 'hello db cli' in result.output

        assert id(db.session) == sess_id



import importlib

import flask
from werkzeug.datastructures import ImmutableDict

import keg.cli
import keg.config
from keg.ctx import KegRequestContext
import keg.logging
import keg.signals as signals
from keg.extensions import lazy_gettext as _
from keg.templating import _keg_default_template_ctx_processor, AssetsExtension
from keg.utils import classproperty, visit_modules, hybridmethod
import keg.web


class KegAppError(Exception):
    pass


class Keg(flask.Flask):
    import_name = None
    use_blueprints = ()
    config_class = keg.config.Config
    logging_class = keg.logging.Logging

    _cli = None
    cli_loader_class = keg.cli.CLILoader

    db_enabled = False
    db_visit_modules = ['.model.entities']
    db_manager = None

    jinja_options = ImmutableDict(
        extensions=[AssetsExtension]
    )

    template_filters = ImmutableDict()
    template_globals = ImmutableDict()

    visit_modules = False

    _init_ran = False

    def __init__(self, import_name=None, *args, **kwargs):

        # flask requires an import name, so we should too.
        if import_name is None and self.import_name is None:
            raise KegAppError(_('Please set the "import_name" attribute on your app class or'
                                ' pass it into the app instance.'))

        # passed in value takes precedence
        import_name = import_name or self.import_name

        self._init_config = kwargs.pop('config', {})

        flask.Flask.__init__(self, import_name, *args, **kwargs)

    def make_config(self, instance_relative=False):
        """
            Needed for Flask <= 0.10.x so we can set the configuration class
            being used.  Once 0.11 comes out, Flask supports setting the config_class on the app.
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        return self.config_class(root_path, self.default_config)

    def init(self, config_profile=None, use_test_profile=False, config=None):
        if self._init_ran:
            raise KegAppError(_('init() already called on this instance'))
        self._init_ran = True

        self.init_config(config_profile, use_test_profile, config)
        self.init_logging()
        self.init_error_handling()
        self.init_extensions()
        self.init_routes()
        self.init_blueprints()
        self.init_jinja()
        self.init_visit_modules()

        self.on_init_complete()
        signals.app_ready.send(self)
        signals.init_complete.send(self)

        # return self for easy chaining, i.e. app = MyKegApp().init()
        return self

    def on_init_complete(self):
        """ For subclasses to override """
        pass

    def init_config(self, config_profile, use_test_profile, config):
        init_config = self._init_config.copy()
        init_config.update(config or {})

        self.config.init_app(config_profile, self.import_name, self.root_path, use_test_profile)

        self.config.update(init_config)

        signals.config_ready.send(self)
        signals.config_complete.send(self)
        self.on_config_complete()

    def on_config_complete(self):
        """ For subclasses to override """
        pass

    def init_extensions(self):
        self.init_db()

        # Keg components are essentially flask extensions, so they need to be set up during the
        # init process. Being extensions, it makes sense to trigger this here. But, they will best
        # be set up at the end of the init process, because other app-specific extensions may be
        # set up in `on_init_complete` for error handling/notifications and the like.
        signals.init_complete.connect(self.init_registered_components, sender=self)

    def db_manager_cls(self):
        from keg.db import DatabaseManager
        return DatabaseManager

    def init_db(self):
        if self.db_enabled:
            cls = self.db_manager_cls()
            self.db_manager = cls(self)

    def init_registered_components(self, app):
        # KEG_REGISTERED_COMPONENTS is presumed to be a set/list/iterable of dotted paths usable
        # for import. At the top level of the imported path, there should be a `__component__`
        # that takes the dotted path that was used for import (as an absolute parent for relative
        # imports) and has an init_app. Ideally, based on KegComponent.
        for comp_path in self.config.get('KEG_REGISTERED_COMPONENTS', set()):
            comp_module = importlib.import_module(comp_path)
            comp_object = getattr(comp_module, '__component__')
            comp_object.init_app(self, parent_path=comp_path)

    def init_blueprints(self):
        for blueprint in self.use_blueprints:
            self.register_blueprint(blueprint)

    def init_logging(self):
        self.logging = self.logging_class(self.config)
        self.logging.init_app()

    def init_error_handling(self):
        # handle status codes
        generic_errors = range(500, 506)
        for err in generic_errors:
            self.errorhandler(err)(self.handle_server_error)

        # utility to abort responses
        self.errorhandler(keg.web.ImmediateResponse)(keg.web.handle_immediate_response)

    def init_jinja(self):
        self.jinja_env.filters.update(self.template_filters)

        # template_context_processors is supposed to be functions that return dictionaries where
        # the key is the name of the template variable and the value is the value.
        # First, add Keg defaults
        self.template_context_processors[None].append(_keg_default_template_ctx_processor)
        self.template_context_processors[None].append(lambda: self.template_globals)

    def init_visit_modules(self):
        if self.visit_modules:
            visit_modules(self.visit_modules, self.import_name)

    def handle_server_error(self, error):
        # send_exception_email()
        return '500 SERVER ERROR<br/><br/>{}'.format(_('administrators notified'))

    def request_context(self, environ):
        """Request context for the app."""
        return KegRequestContext(self, environ)

    def _cli_getter(cls):  # noqa: first argument is not self in this context due to @classproperty
        if cls._cli is None:
            cal = cls.cli_loader_class(cls)
            cls._cli = cal.create_group()
        return cls._cli
    cli = classproperty(_cli_getter, ignore_set=True)

    @classmethod
    def environ_key(cls, key):
        # App names often have periods and it is not possibe to export an
        # environment variable with a period in it.
        name = cls.import_name.replace('.', '_').upper()
        return '{}_{}'.format(name, key.upper())

    @classmethod
    def testing_prep(cls, **config):
        """
            Make sure an instance of this class exists in a state that is ready for testing to
            commence.

            Trigger `signal.testing_run_start` the first time this method is called for an app.
        """
        # For now, do the import here so we don't have a hard dependency on WebTest
        from keg.testing import ContextManager
        if cls is Keg:
            raise TypeError(_('Don\'t use testing_prep() on Keg.  Create a subclass first.'))
        cm = ContextManager.get_for(cls)

        # if the context manager's app isn't ready, that means this will be the first time the app
        # is instantiated.  That seems like a good indicator that tests are just beginning, so it's
        # safe to trigger the signal.  We don't want the signal to fire every time b/c
        # testing_prep() can be called more than once per test run.
        trigger_signal = not cm.is_ready()
        cm.ensure_current(config)

        if trigger_signal:
            signals.testing_run_start.send(cm.app)

        return cm.app

    @property
    def logger(self):
        """Standard logger for the app."""
        return self.logging.app_logger

    @hybridmethod
    def route(self, rule, **options):
        """ Same as Flask.route() and will be used when in an instance context. """
        return super(Keg, self).route(rule, **options)

    @route.classmethod
    def route(cls, rule, **options):  # noqa
        """
        Enable .route() to be used in a class context as well.  E.g.::

            KegApp.route('/something'):
            def view_something():
                pass
        """
        def decorator(f):
            if not hasattr(cls, '_routes'):
                cls._routes = []
            cls._routes.append((f, rule, options))
            return f
        return decorator

    def init_routes(self):
        if not hasattr(self, '_routes'):
            return
        for func, rule, options in self._routes:
            # We follow the same logic here as Flask.route() decorator.
            endpoint = options.pop('endpoint', None)
            self.add_url_rule(rule, endpoint, func, **options)


@signals.init_complete.connect
def flask_wtf_bugfix(app):
    """ Needed until https://github.com/lepture/flask-wtf/issues/301 is fixed & released. """
    if not app.config.get('FIX_FLASK_WTF_BUG', True):
        return

    try:
        import flask_wtf
    except ImportError:
        return

    version = tuple(map(lambda x: int(x), flask_wtf.__version__.split('.')))

    if version < (0, 14, 0):
        return

    @app.teardown_request
    def cleanup_csrf_cache(error):
        if 'csrf_token' in flask.g:
            del flask.g.csrf_token



from __future__ import absolute_import
from __future__ import print_function

import click
import keg
from keg_apps.profile.app import ProfileApp


def cli_entry():
    ProfileApp.cli.main()


@ProfileApp.cli.command('show-profile')
def show_profile():
    click.echo(keg.current_app.config['PROFILE_FROM'])


if __name__ == '__main__':
    cli_entry()



import os
from unittest import mock

import pytest

from keg.cli import dotenv, get_load_dotenv
from keg.testing import CLIBase, app_config
from keg_apps.cli import CLIApp
from keg_apps.cli2.app import CLI2App
from keg_apps.db.app import DBApp


need_dotenv = pytest.mark.skipif(
    not get_load_dotenv() or dotenv is None, reason='dotenv not supported'
)


class TestCLI(CLIBase):
    app_cls = CLIApp
    cmd_name = 'hello'

    def test_class_command(self):
        result = self.invoke()
        assert 'hello keg test' in result.output

    def test_arg_command(self):
        result = self.invoke(cmd_name='foo2')
        assert 'hello foo' in result.output

    def test_argument_passing(self):
        result = self.invoke('bar', cmd_name='foo2')
        assert 'hello bar' in result.output

    def test_missing_command(self):
        result = self.invoke(cmd_name='baz', exit_code=2)
        assert 'No such command \'baz\'' in result.output

    @need_dotenv
    def test_dotenv(self):
        test_dir = os.path.dirname(__file__)
        # ensure flask looks in the expected working directory
        working_dir = os.path.abspath(os.path.join(test_dir, '..', '..'))
        os.chdir(working_dir)

        # Place dotenv file in search path of python-dotenv
        flaskenv = os.path.join(working_dir, '.flaskenv')
        try:
            with open(flaskenv, 'w') as f:
                f.write('FOO=bar')
            assert 'FOO' not in os.environ
            self.invoke()
            assert os.environ['FOO'] == 'bar'
        finally:
            os.remove(flaskenv)
            del os.environ['FOO']

    @need_dotenv
    @mock.patch.dict(os.environ, {'FLASK_SKIP_DOTENV': '1'})
    def test_disable_dotenv_from_env(self):
        test_dir = os.path.dirname(__file__)
        # ensure flask looks in the expected working directory
        working_dir = os.path.abspath(os.path.join(test_dir, '..', '..'))
        os.chdir(working_dir)

        # Place dotenv file in search path of python-dotenv
        flaskenv = os.path.join(working_dir, '.flaskenv')
        try:
            with open(flaskenv, 'w') as f:
                f.write('FOO=bar')
            assert 'FOO' not in os.environ
            self.invoke()
            assert 'FOO' not in os.environ
        finally:
            os.remove(flaskenv)

    @pytest.mark.skipif(True, reason='reminder')
    def test_default_exception_handling(self):
        """
            invoke_command() does some stuff to help with exceptions, it should be tested or
            removed.  So comment in that function for details.
        """


class TestCLI2(CLIBase):
    # We don't set an app_cls (see below), but we do want any current app context cleaned
    # up before constructing a new one.
    force_app_context_cleanup = True

    @classmethod
    def setup_class(cls):
        CLIBase.setup_class()
        # Use this, instead of setting app_cls on the testing class, in order to test that the
        # machinery in place for using the current app is working.
        CLI2App.testing_prep()

    def test_invoke(self):
        result = self.invoke('hello1')
        assert 'hello1' in result.output

    def test_no_commands_help_message(self):
        result = self.invoke()
        assert 'Usage: ' in result.output
        assert '--quiet         Set default logging level to logging.WARNING' in result.output
        assert '--profile TEXT  Name of the configuration profile to use.' in result.output
        assert '--help-all      Show all commands with subcommands.' in result.output
        assert 'develop       Developer info and utils.' in result.output
        assert 'hello1' in result.output

    def test_quiet(self):
        # the asserts for these tests are in keg_apps.cli2.cli
        result = self.invoke('--quiet', 'is-quiet')
        assert 'printed foo' in result.output
        assert 'logged foo' not in result.output

        result = self.invoke('is-not-quiet')
        assert 'printed foo' in result.output
        assert 'logged foo' in result.output

    def test_help_all(self):
        expected_lines = [
            'Usage', '', 'Options', '--profile', '--quiet', '--help-all', '--help',
            'Commands', 'develop', 'Commands', 'config', 'db', 'Commands', 'clear',
            'init', 'routes',
            'run', 'shell', 'templates', 'hello1', 'is-not-quiet', 'is-quiet', 'reverse',
            ''
        ]
        result = self.invoke('--help-all')
        output_lines = result.output.split('\n')
        assert len(expected_lines) == len(output_lines), result.output
        for command, line in zip(expected_lines, output_lines):
            assert line.lstrip().startswith(command), f'"{line}" does not start with "{command}"'

    def test_handle_input(self):
        result = self.invoke('reverse', input='abcde')
        assert 'Input: abcde' in result.output
        assert 'edcba' in result.output


class TestConfigCommand(CLIBase):
    app_cls = CLIApp
    cmd_name = 'develop config'

    def test_output(self):
        result = self.invoke()
        # use a value that should be towards the end of the output as a reasonble indicator the
        # command worked as expected.
        assert 'USE_X_SENDFILE = False' in result.output

    def test_database_disabled(self):
        result = self.invoke(cmd_name='develop db')
        assert 'Database not enabled for this app.  No subcommands available.' in result.output
        assert 'Lists database related sub-commands.' not in result.output

    def test_app_config(self):
        with app_config(FOO_NAME='Bar'):
            result = self.invoke()
        assert 'Bar' in result.output


class TestDatabaseCommands(CLIBase):
    app_cls = DBApp
    cmd_name = 'develop db'

    def test_help_info(self):
        result = self.invoke()
        assert 'Lists database related sub-commands.' in result.output
        assert 'init   Create all db objects, send related events.' in result.output

    @mock.patch('keg.db.DatabaseManager.db_init')
    def test_init(self, m_db_init):
        result = self.invoke('init')
        assert 'Database initialized' in result.output
        m_db_init.assert_called_once_with()

    @mock.patch('keg.db.DatabaseManager.db_init_with_clear')
    def test_init_with_clear(self, m_db_init_wc):
        result = self.invoke('init', '--clear-first', '--yes')
        assert 'Database cleared and initialized' in result.output
        m_db_init_wc.assert_called_once_with()

    @mock.patch('keg.db.DatabaseManager.db_clear')
    def test_clear(self, m_db_clear):
        result = self.invoke('clear', '--yes')
        assert 'Database cleared' in result.output
        m_db_clear.assert_called_once_with()



from __future__ import absolute_import

import flask

from keg.web import BaseView as KegBaseView

blueprint = flask.Blueprint('other', __name__)


class BaseView(KegBaseView):
    blueprint = blueprint


class AutoAssign(BaseView):
    auto_assign = ('bar', 'baz')

    def get(self):
        self.bar = 'bar'
        self.baz = 'baz'
        self.foo = 'foo'


class AutoAssignWithResponse(BaseView):
    auto_assign = ('bar',)
    template_name = 'other/auto-assign.html'

    def get(self):
        self.bar = 'bar'
        return self.render()


class BlankView(BaseView):

    def get(self):
        return ''


class ResponseMiddleware(BaseView):
    url = '/response-middleware/<string:name>'

    def pre_response(self, _response):
        if _response == 'bar':
            return None
        if _response == 'baz':
            return ''
        return _response + '_test'

    def get(self, name):
        return name



from __future__ import absolute_import
from __future__ import print_function

from keg_apps import fix_sys_path
fix_sys_path()

import click  # noqa: E402

from keg.app import Keg  # noqa: E402
from keg_apps.extensions import gettext as _  # noqa: E402


class CLIApp(Keg):
    import_name = __name__
    # Silence the keyring warning messages or CLI output tests get confused b/c of the warning.
    keyring_enabled = False


@CLIApp.cli.command()
def hello():
    print('hello keg test')  # noqa


@CLIApp.cli.command()
@click.argument('name', default='foo')
def foo2(name):
    print((_('hello {name}', name=name)))  # noqa


@CLIApp.cli.command('catch-error')
def catch_error():
    raise Exception(_('deliberate exception for testing'))


@CLIApp.cli.command('verify-translations', help='Verifies all strings marked for translation')
def verify_translations():
    from pathlib import Path
    from morphi.messages.validation import check_translations

    root_path = Path(__file__).resolve().parent.parent
    check_translations(
        root_path,
        'keg',
    )


if __name__ == '__main__':
    CLIApp.cli.main()



import os.path as osp

import appdirs
from blazeutils.helpers import tolist
import flask
from pathlib import PurePath
from werkzeug.utils import (
    import_string,
    ImportStringError
)

from keg.utils import app_environ_get, pymodule_fpaths_to_objects


class ConfigurationError(Exception):
    pass


class SubstituteValue(object):
    def __init__(self, value):
        self.value = value


substitute = SubstituteValue


class Config(flask.Config):
    default_config_locations = [
        # Keg's defaults
        'keg.config.DefaultProfile',
        # Keg's defaults for the selected profile
        'keg.config.{profile}',
        # App defaults for all profiles
        '{app_import_name}.config.DefaultProfile',
        # apply the profile specific defaults that are in the app's config file
        '{app_import_name}.config.{profile}',
    ]

    def from_obj_if_exists(self, obj_location):
        try:
            self.from_object(obj_location)
            self.configs_found.append(obj_location)
        except ImportStringError as e:
            if obj_location not in str(e):
                raise

    def default_config_locations_parsed(self):
        retval = []
        for location in self.default_config_locations:
            # if no profile is given, the location want's one, that location isn't valid
            if '{profile}' in location and self.profile is None:
                continue

            retval.append(location.format(app_import_name=self.app_import_name,
                                          profile=self.profile))
        return retval

    def init_app(self, app_config_profile, app_import_name, app_root_path, use_test_profile,
                 config_file_objs=None):
        self.use_test_profile = use_test_profile
        self.profile = app_config_profile
        self.dirs = appdirs.AppDirs(app_import_name, appauthor=False, multipath=True)
        self.app_import_name = app_import_name
        self.app_root_path = app_root_path
        self.config_paths_unreadable = []

        if config_file_objs:
            self.config_file_objs = config_file_objs
        else:
            self.config_file_objs = []
            possible_config_fpaths = self.config_file_paths()
            fpaths_to_objects = pymodule_fpaths_to_objects(possible_config_fpaths)
            for fpath, objects, exc in fpaths_to_objects:
                if objects is None:
                    self.config_paths_unreadable.append((fpath, exc))
                else:
                    self.config_file_objs.append((fpath, objects))

        if self.profile is None:
            self.profile = self.determine_selected_profile()

        self.configs_found = []

        for dotted_location in self.default_config_locations_parsed():
            dotted_location = dotted_location.format(app_import_name=app_import_name,
                                                     profile=self.profile)
            self.from_obj_if_exists(dotted_location)

        # apply settings from any of this app's configuration files
        for fpath, objects in self.config_file_objs:
            if self.profile in objects:
                self.from_object(objects[self.profile])
                self.configs_found.append('{}:{}'.format(fpath, self.profile))

        sub_values = self.substitution_values()
        self.substitution_apply(sub_values)

    def config_file_paths(self):
        dirs = self.dirs

        config_fname = '{}-config.py'.format(self.app_import_name)

        dpaths = []
        if appdirs.system != 'win32':
            dpaths.extend(dirs.site_config_dir.split(':'))
            dpaths.append('/etc/{}'.format(self.app_import_name))
            dpaths.append('/etc')
        else:
            system_drive = PurePath(dirs.site_config_dir).drive
            system_etc_dir = PurePath(system_drive, '/', 'etc')
            dpaths.extend((
                dirs.site_config_dir,
                system_etc_dir.joinpath(self.app_import_name).__str__(),
                system_etc_dir.__str__()
            ))
        dpaths.append(dirs.user_config_dir)
        dpaths.append(osp.dirname(self.app_root_path))

        fpaths = [osp.join(dpath, config_fname) for dpath in dpaths]

        return fpaths

    def email_error_to(self):
        error_to = self.get('KEG_EMAIL_ERROR_TO')
        override_to = self.get('KEG_EMAIL_OVERRIDE_TO')
        if override_to:
            return tolist(override_to)
        return tolist(error_to)

    def determine_selected_profile(self):
        # if we find the value in the environment, use it
        profile = app_environ_get(self.app_import_name, 'CONFIG_PROFILE')
        if profile is not None:
            return profile

        use_test_profile = app_environ_get(self.app_import_name, 'USE_TEST_PROFILE', '')
        if use_test_profile.strip() or self.use_test_profile:
            return 'TestProfile'

        # look for it in the app's main config file (e.g. myapp.config)
        app_config = import_string('{}.config'.format(self.app_import_name), silent=True)
        if app_config and hasattr(app_config, 'DEFAULT_PROFILE'):
            profile = app_config.DEFAULT_PROFILE

        # Look for it in all the config files found.  This loops from lowest-priority config file
        # to highest priority, so the last file found with a value is kept.  Accordingly, any app
        # specific file has priority over the app's main config file, which could be set just above.
        for fpath, objects in self.config_file_objs:
            if 'DEFAULT_PROFILE' in objects:
                profile = objects['DEFAULT_PROFILE']

        return profile

    def substitution_values(self):
        return dict(
            user_log_dir=self.dirs.user_log_dir,
            app_import_name=self.app_import_name,
        )

    def substitution_apply(self, sub_values):
        for config_key, config_value in self.items():
            if not isinstance(config_value, SubstituteValue):
                continue
            new_value = config_value.value.format(**sub_values)
            self[config_key] = new_value


# The following three classes are default configuration profiles
class DefaultProfile(object):
    KEG_DIR_MODE = 0o777
    KEG_ENDPOINTS = dict(
        home='public.home',
        login='public.home',
        after_login='public.home',
        after_logout='public.home',
    )

    KEG_DB_DIALECT_OPTIONS = {}


class DevProfile(object):
    DEBUG = True


class TestProfile(object):
    DEBUG = True
    TESTING = True
    KEG_LOG_SYSLOG_ENABLED = False

    # set this to allow generation of URLs without a request context
    SERVER_NAME = 'keg.example.com'

    # simple value for testing is fine
    SECRET_KEY = '12345'

    # Sane default values for testing to get rid of warnings.
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'



import datetime as dt
import logging
from logging.handlers import SysLogHandler
import os
import stat

from pythonjsonlogger import jsonlogger

from keg.extensions import lazy_gettext as _

log = logging.getLogger(__name__)


class Logging(object):
    stream_format_str = '%(levelname)s - %(name)s - %(message)s'
    syslog_format_str = '%(levelname)s [%(name)s:%(lineno)d]: %(message)s'
    json_format_str = '%(pathname) %(funcName) %(lineno) %(message) %(levelname)' \
                      ' %(name)s %(process) %(processName) %(message)'

    def __init__(self, config):
        self.config = config
        self.app_logger = logging.getLogger(config.app_import_name)

        self.logging_level = config.get('KEG_LOG_LEVEL', logging.INFO)
        self.logged_locations = config.get('KEG_LOG_MANAGED_LOGGERS', [None])

        self.syslog_enabled = config.get('KEG_LOG_SYSLOG_ENABLED', True)
        self.stream_enabled = config.get('KEG_LOG_STREAM_ENABLED', True)

        if 'KEG_LOG_STDOUT_FORMAT_STR' in config:
            self.stream_format_str = config['KEG_LOG_STDOUT_FORMAT_STR']

        if 'KEG_LOG_SYSLOG_FORMAT_STR' in config:
            self.syslog_format_str = config['KEG_LOG_SYSLOG_FORMAT_STR']

        if 'KEG_LOG_JSON_FORMAT_STR' in config:
            self.json_format_str = config['KEG_LOG_JSON_FORMAT_STR']

        self.syslog_format_json = config.get('KEG_LOG_SYSLOG_JSON', False)

        # Why the ".app"?  Because for a given application, you may be logging to syslog from
        # multiple locations.  For instance, uWSGI might log to syslog for this app.  In that case,
        # all the message's for the app could be namespace like:
        #  - kegdemo.app
        #  - kegdemo.uwsgi
        #  - kegdemo.uwsgi-req
        #  - kegdemo.celery
        #  - etc.
        syslog_default_ident = '{}.app'.format(config.app_import_name)
        self.syslog_ident = config.get('KEG_LOG_SYSLOG_IDENT', syslog_default_ident)

        self.auto_clear_handlers = config.get('KEG_LOG_AUTO_CLEAR_HANDLERS', True)

    def init_app(self):
        self.loggers = [logging.getLogger(location) for location in self.logged_locations]
        self.init_logger_level()
        # If we load multiple instances of the app (usually in testing) we probably don't want
        # multiple handlers created.
        if self.auto_clear_handlers:
            self.clear_keg_handlers()
        if self.stream_enabled:
            self.init_stream()
        if self.syslog_enabled:
            self.init_syslog()

    def init_logger_level(self):
        for logger in self.loggers:
            logger.setLevel(self.logging_level)

    def init_stream(self):
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(self.stream_format_str))
        handler.setLevel(self.logging_level)

        self.add_handler(handler)

    def init_syslog(self):
        handler = self.create_syslog_handler()
        # A space is needed at the end of the ident so syslog recognizes it as the
        # app name and not part of the message.
        handler.ident = self.syslog_ident.rstrip() + ' '
        handler.setFormatter(self.create_syslog_formatter())
        handler.setLevel(self.logging_level)

        self.add_handler(handler)

    def add_handler(self, handler):
        for logger in self.loggers:
            handler._from_keg_logging = True
            logger.addHandler(handler)

    def clear_keg_handlers(self):
        for logger in self.loggers:
            self.clear_logger_handlers(logger)

    def clear_logger_handlers(self, logger):
        keep_handlers = []
        for handler in logger.handlers:
            if hasattr(handler, '_from_keg_logging'):
                handler.flush()
                handler.close()
            else:
                keep_handlers.append(handler)
        logger.handlers = keep_handlers

    def create_syslog_handler(self):
        kwargs = self.config.get('KEG_LOG_SYSLOG_ARGS', {})
        if 'address' not in kwargs:
            address = find_syslog_address(self.config)
            if address:
                log.debug(_('Using syslog address: {address}', address=address))
                kwargs['address'] = address
            else:
                log.debug(_('did not find syslog address'))
        return SysLogHandler(**kwargs)

    def create_syslog_formatter(self):
        if self.syslog_format_json:
            return self.create_json_formatter()
        return logging.Formatter(self.syslog_format_str)

    def create_json_formatter(self):
        prefix = self.config.get('KEG_LOG_SYSLOG_JSON_PREFIX', '@cee:')
        kwargs = self.config.get('KEG_LOG_JSON_FORMATTER_KWARGS', {})
        kwargs.setdefault('prefix', prefix)
        return JSONFormatter(self.json_format_str, **kwargs)


def find_syslog_address(config):
    if _is_socket('/var/run/syslog'):
        return '/var/run/syslog'
    if _is_socket('/dev/log'):
        return '/dev/log'


def _is_socket(path):
    if not os.path.exists(path):
        return False
    mode = os.stat(path).st_mode
    return stat.S_ISSOCK(mode)


class JSONFormatter(jsonlogger.JsonFormatter):
    def process_log_record(self, log_record):
        # Log processing providers like logzio often auto-recognize a field labeled "timestamp".
        log_record['timestamp'] = dt.datetime.utcnow().isoformat()
        return log_record



import inspect
import sys
from warnings import warn

from blazeutils.strings import case_cw2us, case_cw2dash
import flask
from flask import current_app, request
from flask.views import MethodView, http_method_funcs
try:
    from flask.views import MethodViewType
except ImportError:
    MethodViewType = None
from jinja2 import TemplateNotFound
from werkzeug.datastructures import MultiDict

from keg.extensions import lazy_gettext as _
from keg.utils import validate_arguments, ArgumentValidationError


class ImmediateResponse(Exception):
    def __init__(self, response):
        self.response = response


def handle_immediate_response(exc):
    return exc.response


def flash_abort(message):
    flask.flash(message, 'error')
    raise ImmediateResponse


def redirect(endpoint, *args, **kwargs):
    if endpoint.upper() == endpoint:
        keg_endpoints = flask.current_app.config['KEG_ENDPOINTS']
        if endpoint.lower() in keg_endpoints:
            endpoint = keg_endpoints[endpoint.lower()]

    resp = flask.redirect(flask.url_for(endpoint, *args, **kwargs))
    raise ImmediateResponse(resp)


class ViewArgumentError(Exception):
    pass


def _werkzeug_multi_dict_conv(md):
    '''
        Werzeug Multi-Dicts are either flat or lists, but we want a single value
        if only one value or a list if multiple values
    '''
    retval = {}
    for key, value in md.to_dict(flat=False).items():
        if len(value) == 1:
            retval[key] = value[0]
        else:
            retval[key] = value
    return retval


def _call_with_expected_args(view, calling_args, method, method_is_bound=True):
    """ handle argument conversion to what the method accepts """
    if isinstance(method, str):
        if not hasattr(view, method):
            return
        method = getattr(view, method)

    try:
        # validate_arguments is made for a function, not a class method
        # so we need to "trick" it by sending self here, but then
        # removing it before the bound method is called below
        pos_args = (view,) if method_is_bound else tuple()
        args, kwargs = validate_arguments(method, pos_args, calling_args.copy())
    except ArgumentValidationError as e:
        msg = _('Argument mismatch occured: method={method}, missing={missing}, '
                'extra_keys={extra_keys}, extra_pos={extra_pos}.'
                '  Arguments available: {calling_args}',
                method=method, missing=e.missing, extra_keys=e.extra,
                extra_pos=e.extra_positional, calling_args=calling_args)
        raise ViewArgumentError(msg)
    if method_is_bound:
        # remove "self" from args since its a bound method
        args = args[1:]
    return method(*args, **kwargs)


class _OldViewMeta(MethodViewType or object):
    def __init__(cls, name, bases, d):
        MethodViewType.__init__(cls, name, bases, d)

        # Assuming child views will always have a blueprint OR they are intended to be used like
        # abstract classes and will never be routed to directly. Apps/libs will need to call
        # assign_blueprint directly if a blueprint is not set in the view class definition, but
        # the view is intended to receive routes.
        if cls.blueprint is not None:
            cls.assign_blueprint(cls.blueprint)


_ViewMeta = _OldViewMeta if MethodViewType is not None else type


class BaseView(MethodView, metaclass=_ViewMeta):
    """
    Base class for all Keg views to inherit from. `BaseView` automatically calculates and installs
    routing, templating, and responding methods for HTTP verb named functions.

    .. literalinclude:: /examples/BaseView.py
    """
    blueprint = None
    url = None
    template_name = None
    _index_method = None

    require_authentication = False
    auto_assign = tuple()
    # names of qs arguments that should be merged w/ URL arguments and passed to view methods
    expected_qs_args = []

    def __init_subclass__(cls, **kwargs):
        """Flask before 2.2.0 used a metaclass to perform view setup, but this
        changed to using `init_subclass`. If the old way is enabled, no need to
        do anything but call the super here."""
        super().__init_subclass__(**kwargs)

        if MethodViewType is None and cls.blueprint is not None:
            cls.assign_blueprint(cls.blueprint)

    def __init__(self, responding_method=None):
        self.responding_method = responding_method

    def calc_responding_method(self):
        if self.responding_method:
            method_name = self.responding_method

            # sometimes the real method name is based on the name of the method combined with the
            # HTTP method/verb.
            method_name_verb = '{}_{}'.format(method_name, request.method.lower())
            if not hasattr(self, method_name) and hasattr(self, method_name_verb):
                method_name = method_name_verb
        else:
            method_name = request.method.lower()

            # If the request method is HEAD and we don't have a handler for it
            # retry with GET.
            if request.method == 'HEAD' and not hasattr(self, 'head'):
                method_name = 'get'

        method_obj = getattr(self, method_name, None)

        assert method_obj is not None, _('Unimplemented method {name}', name=method_name)

        return method_obj

    def dispatch_request(self, **kwargs):
        self.template_args = {}
        calling_args = self.process_calling_args(kwargs)
        self._calling_args = calling_args
        _call_with_expected_args(self, calling_args, 'pre_auth')
        _call_with_expected_args(self, calling_args, 'check_auth')
        _call_with_expected_args(self, calling_args, 'pre_loaders')
        self.call_loaders(calling_args)
        _call_with_expected_args(self, calling_args, 'pre_method')

        method_obj = self.calc_responding_method()
        response = _call_with_expected_args(self, calling_args, method_obj)

        # The following condition is intended to guard against None or no return
        # from the responding method. But, an empty string is a valid response that
        # would fail the falsy test, so check that one specifically.
        if not response and response != '':
            self.process_auto_assign()
            _call_with_expected_args(self, calling_args, 'pre_render')
            response = self.render()
        calling_args['_response'] = response
        pre_response = _call_with_expected_args(self, calling_args, 'pre_response')
        if pre_response is not None:
            return pre_response
        return response

    def process_calling_args(self, urlargs):
        # start with query string arguments that are expected
        args = MultiDict()
        if self.expected_qs_args:
            for k in request.args.keys():
                if k in self.expected_qs_args:
                    args.setlist(k, request.args.getlist(k))

        # add URL arguments, replacing GET arguments if they are there.  URL
        # arguments get precedence and we don't want to just .update()
        # because that would allow arbitrary get arguments to affect the
        # values of the URL arguments
        for k, v in urlargs.items():
            args[k] = v

        return _werkzeug_multi_dict_conv(args)

    def call_loaders(self, calling_args):
        for attr_name, method_obj in inspect.getmembers(self, inspect.ismethod):
            if not attr_name.endswith('_loader'):
                continue
            arg_key = attr_name[:-7]
            retval = _call_with_expected_args(self, calling_args, method_obj)
            if retval is None:
                flask.abort(404)
            # have to add it to internal variable b/c calling_args is a copy, not the actual object
            self._calling_args[arg_key] = retval

    def check_auth(self):
        pass
        # if self.require_authentication and not current_user.is_authenticated():
        #    flask.abort(401)

    def process_auto_assign(self):
        for key in self.auto_assign:
            self.assign(key, getattr(self, key))

    def calc_class_fname(self, use_us=False):
        if use_us:
            return case_cw2us(self.__class__.__name__)
        return case_cw2dash(self.__class__.__name__)

    def calc_template_name(self, use_us=False):
        if self.template_name is not None:
            return self.template_name
        template_path = '{}.html'.format(self.calc_class_fname(use_us=use_us))
        blueprint_name = request.blueprint
        if blueprint_name:
            template_path = '{}/{}'.format(blueprint_name, template_path)

        # This block and `use_us` usage can be removed when we drop
        # template names generated with underscores.
        jinja_env = current_app.jinja_env
        try:
            jinja_env.get_template(template_path)
        except TemplateNotFound:
            raise_original_exception = False
            if not use_us:
                # Fall back to underscore-named templates for now, but warn if
                # that succeeds
                try:
                    template_path = self.calc_template_name(use_us=True)
                    warn(
                        'Templates named by underscore-notated class names are '
                        'deprecated and will not be supported. Rename the template '
                        'files using dashes.',
                        DeprecationWarning,
                        stacklevel=2
                    )
                except TemplateNotFound:
                    # Neither template exists. Raise the exception from the dashed name
                    raise_original_exception = True
            if use_us or raise_original_exception:
                raise

        return template_path

    def assign(self, key, value):
        self.template_args[key] = value

    def render(self):
        return flask.render_template(self.calc_template_name(), **self.template_args)

    @classmethod
    def calc_url(cls, use_blueprint=True):
        # calc_url will generally return the full url to reach a view. Turn use_blueprint off when
        # setting up the blueprint itself (see init_routes and init_blueprint)
        prefix = (cls.blueprint.url_prefix or '') if cls.blueprint and use_blueprint else ''
        if cls.url is not None:
            return prefix + cls.url
        return prefix + '/' + case_cw2dash(cls.__name__)

    @classmethod
    def calc_endpoint(cls, use_blueprint=True):
        # calc_endpoint will generally return the full endpoint for a view. Turn use_blueprint off
        # when setting up the blueprint itself (see init_routes and init_blueprint)
        prefix = (cls.blueprint.name + '.') if cls.blueprint and use_blueprint else ''
        return prefix + case_cw2dash(cls.__name__)

    @classmethod
    def init_routes(cls):
        cls.url_rules = []
        cls.view_funcs = {}

        def method_with_rules(obj):
            # isroutine() matches functions (PY3) and unbound methods (PY2).
            return inspect.isroutine(obj) and hasattr(obj, '_keg_rules')

        for method_name, method_obj in inspect.getmembers(cls, predicate=method_with_rules):
            for rule, options in method_obj._keg_rules:
                method_route = MethodRoute(method_name, rule, options,
                                           cls.calc_url(use_blueprint=False),
                                           cls.calc_endpoint(use_blueprint=False))
                mr_options = method_route.options()
                if method_route.endpoint not in cls.view_funcs:
                    view_func = cls.as_view(method_route.view_func_name,
                                            method_route.sanitized_method_name('_'))
                    cls.view_funcs[method_route.endpoint] = view_func
                mr_options['view_func'] = cls.view_funcs[method_route.endpoint]
                cls.blueprint.add_url_rule(method_route.rule(), **mr_options)
            if method_name in http_method_funcs and cls.methods:
                # A @route() is being used on a method with the same name as an HTTP verb.
                # Since this method is being explicitly routed, the automatic rule that would
                # be created due to MethodView logic should not apply.
                cls.methods.remove(method_name.upper())

    @classmethod
    def init_blueprint(cls, rules):
        endpoint = cls.calc_endpoint(use_blueprint=False)
        class_url = cls.calc_url(use_blueprint=False)

        if cls.methods:
            # Flask assigns the endpoint to the __name__ atribute of the function it creates and
            # therefore the endpoint must be a `str`.
            str_endpoint = str(endpoint)
            view_func = cls.as_view(str_endpoint)

            for rule, options in rules:
                if rule and rule.startswith('/'):
                    class_url = rule
                elif not rule:
                    rule = class_url
                else:
                    rule = '{}/{}'.format(class_url, rule)
                cls.blueprint.add_url_rule(rule, endpoint=endpoint, view_func=view_func, **options)

            if not rules:
                cls.blueprint.add_url_rule(class_url, endpoint=endpoint, view_func=view_func)

        for rule, options in cls.url_rules:
            cls.blueprint.add_url_rule(rule, **options)

    @classmethod
    def assign_blueprint(cls, blueprint):
        if not blueprint:
            raise Exception('blueprint {} could not be assigned'.format(blueprint))

        cls.blueprint = blueprint
        rules = []
        if hasattr(cls, '_rules'):
            rules = cls._rules
            del cls._rules

        cls.init_routes()
        cls.init_blueprint(rules)


class MethodRoute(object):

    def __init__(self, method_name, rule, options, parent_url, parent_endpoint):
        self.method_name = method_name
        self._rule = rule
        # We use destructive operations on options, so make a copy.
        self._options = options.copy()
        self.parent_url = parent_url
        self.parent_endpoint = parent_endpoint

    def sanitized_method_name(self, separator='-'):
        method_identifier = self.method_name.replace('_', separator)

        # method names can be given like foo_get() and foo_post() which indicate the HTTP verb
        # used but shouldn't affect the URL.
        parts = method_identifier.split(separator)
        method_name_suffix = parts[-1]
        if len(parts) > 1 and method_name_suffix in http_method_funcs:
            last_position = (len(method_name_suffix) + 1) * -1
            return method_identifier[:last_position]

        return method_identifier

    @property
    def endpoint(self):
        return '{}:{}'.format(self.parent_endpoint, self.sanitized_method_name())

    @property
    def view_func_name(self):
        return str('{}_{}'.format(self.parent_endpoint, self.sanitized_method_name('_')))

    def rule(self):
        method_url = '{}/{}'.format(self.parent_url, self.sanitized_method_name())

        # Handle no rule given by @route().
        if not self._rule:
            return method_url

        # Handle a relative rule given by @route().
        if not self._rule.startswith('/'):
            return '{}/{}'.format(method_url, self._rule)

        return self._rule

    def options(self):
        self._options['endpoint'] = self.endpoint
        return self._options


def _methods_calc(get, post, post_only, methods):
    if methods is None:
        methods = set()
    else:
        methods = set(methods)

    if get and not post_only:
        methods.add('GET')
    if post or post_only:
        methods.add('POST')

    return methods


def route(rule=None, get=True, post=False, post_only=False, methods=None, **options):
    options['methods'] = _methods_calc(get, post, post_only, methods)

    def wrapper(func):
        if not hasattr(func, '_keg_rules'):
            func._keg_rules = []
        func._keg_rules.append((rule, options))
        return func

    return wrapper


def rule(rule=None, get=True, post=False, post_only=False, methods=None, **options):
    parent_locals = sys._getframe(1).f_locals
    rules = parent_locals.setdefault('_rules', [])
    options['methods'] = _methods_calc(get, post, post_only, methods)
    rules.append((rule, options))



import contextlib

import click
import click.testing
import flask
from flask_webtest import TestApp
from werkzeug.datastructures import ImmutableMultiDict, MultiDict

from keg import current_app, signals
from keg.utils import app_environ_get


def _config_profile(appcls):
    config_profile = 'TestProfile'
    if appcls.import_name is not None:
        config_profile = app_environ_get(appcls.import_name, 'CONFIG_PROFILE', config_profile)
    return config_profile


class ContextManager(object):
    """
        Facilitates having a single instance of an application ready for testing.

        By default, this is used in ``Keg.testing_prep``.

        Constructor arg is the Keg app class to manage for tests.
    """
    apps = {}

    def __init__(self, appcls):
        self.appcls = appcls
        self.app = None
        self.ctx = None

    def ensure_current(self, config):
        """Ensure the manager's app has an instance set as flask's ``current_app``"""

        if not self.app:
            self.app = self.appcls().init(use_test_profile=True, config=config)
            self.ctx = self.app.app_context()
            self.ctx.push()

        if current_app._get_current_object is not self.app:
            self.ctx.push()

        return self.app

    def cleanup(self):
        """Pop the app context"""
        self.ctx.pop()

    def is_ready(self):
        """Indicates the manager's app instance exists.

        The instance should be created with ``get_for``. Only one ContextManager instance will get
        created in a Python process for any given app. But, ``get_for`` may be called multiple
        times. The first call to ``ensure_current`` will set up the application and bring the
        manager to a ready state."""
        return self.app is not None

    @classmethod
    def get_for(cls, appcls):
        """
            Return the ContextManager instance for the given app class.  Only one ContextManager
            instance will be created in a Python process for any given app.
        """
        if appcls not in cls.apps:
            cls.apps[appcls] = cls(appcls)
        return cls.apps[appcls]


@contextlib.contextmanager
def app_config(**kwargs):
    """
        Set config values on any apps instantiated while the context manager is active.
        This is intended to be used with cli tests where the ``current_app`` in the test will be
        different from the ``current_app`` when the CLI command is invoked, making it very difficult
        to dynamically set app config variables using mock.patch.dict like we normally would.

        Example::

            class TestCLI(CLIBase):
                app_cls = MyApp
                def test_it(self):
                    with testing.app_config(FOO_NAME='Bar'):
                        result = self.invoke('echo-foo-name')
                    assert 'Bar' in result.output
    """
    @signals.config_complete.connect
    def set_config(app):
        app.config.update(kwargs)

    yield


@contextlib.contextmanager
def inrequest(*req_args, args_modifier=None, **req_kwargs):
    """A decorator/context manager to add the flask request context to a test function.

    Allows test to assume a request context without running a full view stack. Use for
    unit-testing a view instance without setting up a webtest instance for the app and
    running requests.

    Flask's ``request.args`` is normally immutable, but in test cases, it can be helpful to
    patch in args without needing to construct the URL. But, we don't want to leave them
    mutable, because potential app bugs could be masked in doing so. To modify args, pass
    in a callable as ``args_modifier`` that takes the args dict to be modified in-place. Args
    will only be mutable for executing the modifier, then returned to immutable for the
    remainder of the scope.

    Assumes that ``flask.current_app`` is pointing to the desired app.

    Example::

        @inrequest('/mypath?foo=bar&baz=boo')
        def test_in_request_args(self):
            assert flask.request.args['foo'] == 'bar'

        def test_request_args_mutated(self):
            def args_modifier(args_dict):
                args_dict['baz'] = 'custom-value'

            with inrequest('/mypath?foo=bar&baz=boo', args_modifier=args_modifier):
                assert flask.request.args['foo'] == 'bar'
                assert flask.request.args['baz'] == 'custom-value'
    """
    with flask.current_app.test_request_context(*req_args, **req_kwargs):
        if callable(args_modifier):
            # Temporarily turn args into a mutable MultiDict to be patched. Then, we must
            # be sure to turn them back immutable, or else tests may end up not catching
            # bugs that attempt to modify request args improperly.
            new_args = MultiDict(flask.request.args)
            args_modifier(new_args)
            flask.request.args = ImmutableMultiDict(new_args)

        yield


def invoke_command(app_cls, *args, **kwargs):
    """Invoke a command using a CLI runner and return the result.

    Optional kwargs:
    - exit_code: Default 0. Process exit code to assert.
    - runner: Default ``click.testing.CliRunner()``. CLI runner instance to use for invocation.
    - use_test_profile: Default True. Drive invoked app to use test profile instead of default.
    """
    exit_code = kwargs.pop('exit_code', 0)
    runner = kwargs.pop('runner', None) or click.testing.CliRunner()
    use_test_profile = kwargs.pop('use_test_profile', True)
    env = kwargs.pop('env', {})

    if use_test_profile:
        app_key = app_cls.environ_key('USE_TEST_PROFILE')
        env[app_key] = 'true'
    result = runner.invoke(app_cls.cli, args, env=env, catch_exceptions=False, **kwargs)

    error_message = 'Command exit code {}, expected {}.  Result output follows:\n{}'
    assert result.exit_code == exit_code, error_message.format(result.exit_code, exit_code,
                                                               result.output)
    return result


def cleanup_app_contexts():
    while flask.current_app:
        cm = ContextManager.get_for(flask.current_app.__class__)
        if cm.ctx:
            cm.cleanup()
        else:
            break

    # support older flask as well
    if flask.current_app and getattr(flask, '_app_ctx_stack'):
        while flask._app_ctx_stack.pop():
            pass


class CLIBase(object):
    """Test class base for testing Keg click commands.

    Creates a CLI runner instance, and allows subclass to call ``self.invoke`` with
    command args.

    Class attributes:
    - app_cls: Optional, will default to ``flask.current_app`` class.
    - cmd_name: Optional, provides default in ``self.invoke`` for ``cmd_name`` kwarg.
    """
    # child classes will need to set these values in order to use the class
    app_cls = None
    cmd_name = None
    force_app_context_cleanup = False

    @classmethod
    def setup_class(cls):
        # If a current app context is set, it may complicate what click is doing to
        # set up and run a specific app.
        if cls.app_cls or cls.force_app_context_cleanup:
            cleanup_app_contexts()

        cls.runner = click.testing.CliRunner()

    def invoke(self, *args, **kwargs):
        """Run a command, perform some assertions, and return the result for testing."""
        cmd_name = kwargs.pop('cmd_name', self.cmd_name)
        if cmd_name is None:
            cmd_name_args = []
        else:
            cmd_name_args = cmd_name.split(' ')
        invoke_args = cmd_name_args + list(args)
        kwargs['runner'] = self.runner

        # If the app_cls isn't set, use the class of the current app
        app_cls = self.app_cls or flask.current_app._get_current_object().__class__

        return invoke_command(app_cls, *invoke_args, **kwargs)


class WebBase(object):
    db = None
    appcls = None

    @classmethod
    def setup_class(cls):
        cls.app = cls.appcls.testing_prep()
        cls.testapp = TestApp(flask.current_app)



# flask.globals.request_ctx is only available in Flask >= 2.2.0
try:
    from flask.globals import request_ctx
except ImportError:
    from flask.globals import _request_ctx_stack
    request_ctx = None

from keg.extensions import lazy_gettext as _


def _get_bc_request_context():
    """Flask 2.2 changed the method of fetching the request context
    from globals. Flask 2.3 will remove the old way of doing it.
    Support both here."""
    if request_ctx is None:
        return _request_ctx_stack.top
    return request_ctx


def _keg_default_template_ctx_processor():
    """Default template context processor.  Injects `assets`.
    """
    reqctx = _get_bc_request_context()
    rv = {}
    if reqctx:
        rv['assets'] = reqctx.assets
    return rv

from jinja2 import nodes  # noqa
from jinja2.ext import Extension  # noqa


class AssetsExtension(Extension):
    # a set of names that trigger the extension.
    tags = set(['assets_include', 'assets_content'])

    def __init__(self, environment):
        super(AssetsExtension, self).__init__(environment)

    def parse(self, parser):
        lineno = parser.stream.current.lineno
        stream = parser.stream

        if stream.current.value == 'assets_include':
            return self.parse_include(parser, stream, lineno)
        return self.parse_content(parser, stream, lineno)

    def parse_include(self, parser, stream, lineno):

        # did asset_include have parameters?
        if stream.look().type == 'block_end':
            args = [nodes.Const(parser.name, lineno=lineno)]
        else:
            # print parser.parse_expression()
            parser.fail(_('asset_include does not yet support parameters'))

        # parse the closing bracket for this asset_include call
        next(stream)
        assert stream.current.type == 'block_end'

        # now return a `CallBlock` node that calls our _import_support
        # helper method on this extension.
        return nodes.CallBlock(self.call_method('_include_support', args),
                               [], [], []).set_lineno(lineno)

    def _include_support(self, template_name, caller):
        """Helper callback."""
        ctx = _get_bc_request_context()
        ctx.assets.load_related(template_name)

        # have to return empty string to avoid exception about None not being iterable.
        return ''

    def parse_content(self, parser, stream, lineno):
        # move from the current Name token to the next token in the stream, which should be the
        # first argument to asset_content tag.
        next(stream)

        # parse the expression
        args = [parser.parse_expression()]

        # now return a `CallBlock` node that calls our support
        # helper method on this extension.
        return nodes.CallBlock(self.call_method('_content_support', args),
                               [], [], []).set_lineno(lineno)

    def _content_support(self, asset_type, caller):
        """Helper callback."""
        ctx = _get_bc_request_context()
        return ctx.assets.combine_content(asset_type)



from unittest import mock

import flask
import pytest

from keg.testing import WebBase
from keg.web import BaseView

from keg_apps.web.app import WebApp


class TestViewCalculations:
    def test_view_defaults_no_blueprint(self):
        class MyView(BaseView):
            pass

        assert MyView.calc_url() == '/my-view'
        assert MyView.calc_endpoint() == 'my-view'

    def test_view_url_no_blueprint(self):
        class MyView(BaseView):
            url = '/some-other-place'

        assert MyView.calc_url() == '/some-other-place'

    def test_view_url_with_blueprint_prefix(self):
        class MyView(BaseView):
            blueprint = flask.Blueprint('testallthethings', __name__, url_prefix='/foo')
            url = '/some-other-place'

        assert MyView.calc_url() == '/foo/some-other-place'
        assert MyView.calc_url(use_blueprint=False) == '/some-other-place'

    def test_view_defaults_with_blueprint(self):
        class MyView(BaseView):
            blueprint = flask.Blueprint('testallthethings', __name__)

        assert MyView.calc_url() == '/my-view'
        assert MyView.calc_endpoint() == 'testallthethings.my-view'
        assert MyView.calc_url(use_blueprint=False) == '/my-view'
        assert MyView.calc_endpoint(use_blueprint=False) == 'my-view'


class TestViewRouting(WebBase):
    appcls = WebApp

    def test_route_endpoints(self):
        all_endpoints = sorted(self.app.url_map._rules_by_endpoint.keys())
        routing_endpoints = list(filter(lambda x: x.startswith('routing.'), all_endpoints))
        assert routing_endpoints == [
            'routing.cars:create',
            'routing.cars:delete',
            'routing.cars:edit',
            'routing.cars:list',
            'routing.explicit-route',
            'routing.hello-req',
            'routing.hello-req2',
            'routing.hello-world',
            'routing.hw-rule-default',
            'routing.misc',
            'routing.misc:an-abs-url',
            'routing.misc:two-routes',
            'routing.planes:list',
            'routing.tickets',
            'routing.trucks:list',
            'routing.verb-routing',
            'routing.verb-routing-sub',
        ]

    def test_verb_routing(self):
        resp = self.testapp.get('/verb-routing')
        assert resp.text == 'method get'

        resp = self.testapp.post('/verb-routing')
        assert resp.text == 'method post'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.verb-routing'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/verb-routing'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.verb-routing'

    def test_subclassing(self):
        resp = self.testapp.get('/verb-routing-sub')
        assert resp.text == 'method get'

        resp = self.testapp.post('/verb-routing-sub')
        assert resp.text == 'method post'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.verb-routing-sub'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/verb-routing-sub'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.verb-routing-sub'

    def test_explicit_route(self):
        resp = self.testapp.get('/some-route')
        assert resp.text == 'get some-route'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.explicit-route'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/some-route'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.explicit-route'

    def test_explicit_route_assigned_blueprint(self):
        resp = self.testapp.get('/some-route-alt', status=404)

        from keg_apps.web.views.routing import ExplicitRouteAlt
        blueprint = flask.Blueprint('test_explicit_route_assigned_blueprint', __name__)

        with pytest.raises(Exception, match='could not be assigned'):
            ExplicitRouteAlt.assign_blueprint(None)

        ExplicitRouteAlt.assign_blueprint(blueprint)
        with mock.patch('flask.current_app._got_first_request', False):
            flask.current_app.register_blueprint(blueprint)

        resp = self.testapp.get('/some-route-alt')
        assert resp.text == 'get some-route alt'

        rules = list(self.app.url_map.iter_rules(
            endpoint='test_explicit_route_assigned_blueprint.explicit-route-alt'
        ))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/some-route-alt'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'test_explicit_route_assigned_blueprint.explicit-route-alt'

    def test_blueprint_routes(self):
        self.testapp.get('/blueprint-test', status=404)
        self.testapp.get('/tanagra/blueprint-test')
        self.testapp.get('/custom-route', status=404)
        self.testapp.get('/tanagra/custom-route')

        from keg_apps.web.views import custom
        assert custom.BlueprintTest.calc_url() == '/tanagra/blueprint-test'
        assert custom.BlueprintTest.calc_url(use_blueprint=False) == '/blueprint-test'
        assert custom.BlueprintTest.calc_endpoint() == 'custom.blueprint-test'
        assert custom.BlueprintTest.calc_endpoint(use_blueprint=False) == 'blueprint-test'
        assert custom.BlueprintTest2.calc_url() == '/tanagra/custom-route'
        assert custom.BlueprintTest2.calc_url(use_blueprint=False) == '/custom-route'
        assert custom.BlueprintTest2.calc_endpoint() == 'custom.blueprint-test2'
        assert custom.BlueprintTest2.calc_endpoint(use_blueprint=False) == 'blueprint-test2'

    def test_hello_world(self):
        resp = self.testapp.get('/hello-world')
        assert resp.text == 'Hello World'

        resp = self.testapp.get('/hello-world/sam')
        assert resp.text == 'Hello sam'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.hello-world'))
        assert len(rules) == 2

        rule = rules.pop()
        assert rule.rule == '/hello-world'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hello-world'

        rule = rules.pop()
        assert rule.rule == '/hello-world/<name>'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hello-world'

    def test_hello_world_defaults(self):
        resp = self.testapp.get('/hwrd')
        assert resp.text == 'Hello RD'

        resp = self.testapp.get('/hwrd/sam')
        assert resp.text == 'Hello sam'

        resp = self.testapp.post('/hwrd')
        assert resp.text == 'post Hello RD'

        resp = self.testapp.post('/hwrd/sam')
        assert resp.text == 'post Hello sam'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.hw-rule-default'))
        assert len(rules) == 2

        rule = rules.pop()
        assert rule.rule == '/hwrd/<name>'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hw-rule-default'

        rule = rules.pop()
        assert rule.rule == '/hwrd'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hw-rule-default'

    def test_route_with_required_rule(self):
        self.testapp.get('/hello-req-opt1', status=404)

        resp = self.testapp.get('/hello-req/foo')
        assert resp.text == 'Hello foo'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.hello-req'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/hello-req/<name>'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hello-req'

    def test_route_no_absolute_single_endpoint(self):
        self.testapp.get('/hello-req2', status=404)

        resp = self.testapp.get('/hello-req2/foo')
        assert resp.text == 'Hello foo'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.hello-req2'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/hello-req2/<name>'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.hello-req2'

    def test_route_plain(self):
        resp = self.testapp.get('/cars/list')
        assert resp.text == 'list'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.cars:list'))
        assert len(rules) == 1
        rule = rules.pop()

        assert rule.rule == '/cars/list'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.cars:list'

    def test_route_method_verb_suffix(self):
        resp = self.testapp.get('/cars/create')
        assert resp.text == 'create get'

        resp = self.testapp.post('/cars/create')
        assert resp.text == 'create post'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.cars:create'))
        assert len(rules) == 2

        rule = rules.pop()
        assert rule.rule == '/cars/create'
        assert rule.methods == {'POST', 'OPTIONS'}
        assert rule.endpoint == 'routing.cars:create'

        rule = rules.pop()
        assert rule.rule == '/cars/create'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.cars:create'

    def test_route_relative(self):
        resp = self.testapp.get('/cars/edit', status=404)

        resp = self.testapp.get('/cars/edit/12')
        assert resp.text == 'form w/ data: 12'

        resp = self.testapp.post('/cars/edit/7')
        assert resp.text == 'update car: 7'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.cars:edit'))
        assert len(rules) == 1

        rule = rules.pop()
        assert rule.rule == '/cars/edit/<int:car_id>'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.cars:edit'

    def test_route_on_http_verb_method(self):
        resp = self.testapp.get('/cars/delete/7')
        assert resp.text == 'delete: 7'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.cars:delete'))
        assert len(rules) == 1

        rule = rules.pop()
        assert rule.rule == '/cars/delete/<int:ident>'
        assert rule.methods == {'GET', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.cars:delete'

    def test_rest_example(self):
        resp = self.testapp.get('/tickets')
        assert resp.text == 'list'

        resp = self.testapp.get('/tickets/5')
        assert resp.text == 'single'

        resp = self.testapp.post('/tickets')
        assert resp.text == 'create'

        resp = self.testapp.put('/tickets/6')
        assert resp.text == 'update'

        resp = self.testapp.patch('/tickets/6')
        assert resp.text == 'partial update'

        resp = self.testapp.delete('/tickets/6')
        assert resp.text == 'delete'

        rules = list(self.app.url_map.iter_rules(endpoint='routing.tickets'))
        assert len(rules) == 2

        rule = rules.pop()
        assert rule.rule == '/tickets'
        assert rule.methods == {'GET', 'POST', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.tickets'

        rule = rules.pop()
        assert rule.rule == '/tickets/<int:ticket_id>'
        assert rule.methods == {'GET', 'PUT', 'PATCH', 'DELETE', 'HEAD', 'OPTIONS'}
        assert rule.endpoint == 'routing.tickets'

    def test_multiple_class_rules(self):
        resp = self.testapp.get('/misc')
        assert resp.text == 'get'

        resp = self.testapp.get('/misc/foo')
        assert resp.text == 'get'

        resp = self.testapp.get('/misc2')
        assert resp.text == 'get'

        resp = self.testapp.get('/misc2/bar', status=405)

        resp = self.testapp.post('/misc2/bar')
        assert resp.text == 'post'

    def test_abs_route(self):
        resp = self.testapp.get('/an-abs-url')
        assert resp.text == 'found me'

    def test_two_routes(self):
        resp = self.testapp.get('/misc/two-routes/8')
        assert resp.text == '17'
        resp = self.testapp.get('/misc/two-routes/9')
        assert resp.text == '17'

    def test_abstract_class_usage(self):
        resp = self.testapp.get('/trucks/list')
        assert resp.text == 'listing Trucks'

        resp = self.testapp.get('/planes/list')
        assert resp.text == 'listing Planes'

    def test_routing_decorator_class_context(self):
        self.testapp.get('/simple', status=200)

        # Make sure options dict gets passed through correctly by making sure we can POST.
        self.testapp.post('/simple', status=200)

        # make sure the endpoints got set correctly
        assert self.app.view_functions['simple1']
        assert self.app.view_functions['simple2']

    def test_routing_decorator_instance_context(self):
        self.testapp.get('/simple3', status=200)



from __future__ import absolute_import

from keg.app import Keg

from keg_apps.extensions import lazy_gettext as _


class DBApp(Keg):
    import_name = 'keg_apps.db'
    db_enabled = True


@DBApp.cli.command()
def hello():
    print(_('hello db cli'))


if __name__ == '__main__':
    DBApp.cli.main()



from __future__ import absolute_import
from __future__ import print_function

from keg_apps.web.app import WebApp


def cli_entry():
    WebApp.cli.main()


if __name__ == '__main__':
    cli_entry()



import logging

import click

from .app import CLI2App

log = logging.getLogger(__name__)


@CLI2App.cli.command('hello1')
def hello1():
    click.echo('hello1')


@CLI2App.cli.command('is-quiet')
def is_quiet():
    print('printed foo')
    log.info('logged foo')


@CLI2App.cli.command('is-not-quiet')
def is_not_quiet():
    print('printed foo')
    log.info('logged foo')


@CLI2App.cli.command()
def reverse():
    data = click.prompt('Input')
    click.echo(data[::-1])



import os

from jinja2 import TemplateSyntaxError
import pytest

from keg import current_app
from keg_apps.templating.app import TemplatingApp


def setup_module(module):
    TemplatingApp.testing_prep()


class TestAssetsInclude(object):
    def render(self, filename):
        template = current_app.jinja_env.get_template(filename)
        return template.render()

    def setup_method(self, method):
        self.ctx = current_app.test_request_context()
        self.ctx.push()
        self.assets = self.ctx.assets

    def teardown_method(self, method):
        self.ctx.pop()

    def test_include(self):
        resp = self.render('assets_in_template.html')
        assert resp.strip() == ''

        assert len(self.assets.content['js']) == 1
        asset_name, filepath, content = self.assets.content['js'].pop()
        assert asset_name == 'assets_in_template.js'
        assert filepath.endswith(
            os.path.join('keg_apps', 'templating', 'templates', 'assets_in_template.js')
        )
        assert content.strip() == '//assets_in_template js file'

        assert len(self.assets.content['css']) == 1
        asset_name, filepath, content = self.assets.content['css'].pop()
        assert asset_name == 'assets_in_template.css'
        assert filepath.endswith(
            os.path.join('keg_apps', 'templating', 'templates', 'assets_in_template.css')
        )
        assert content.strip() == '/* assets_in_template css file */'

    def test_include_with_params(self):
        with pytest.raises(TemplateSyntaxError) as e:
            self.render('assets_with_params.html')
            assert str(e) == 'asset_include does not yet support parameters'

    def test_assets_content(self):
        self.ctx.assets.content['css'].append(('fake-file.css', 'foo', '//fake-file.css content'))
        self.ctx.assets.content['js'].append(('fake-file.js', 'foo', '//fake-file.js content'))
        resp = self.render('assets_content.html')
        lines = resp.splitlines()
        assert lines[0] == '/********************* asset: fake-file.css *********************/'
        # skip newline that combine_content() adds
        assert lines[2] == '//fake-file.css content'
        assert lines[3] == '/********************* asset: fake-file.js *********************/'
        # skip newline that combine_content() adds
        assert lines[5] == '//fake-file.js content'



from .routing import blueprint as routing_bp
from .templating import blueprint as blueprint_bp
from .other import blueprint as other_bp
from .custom import blueprint as custom_bp

all_blueprints = [routing_bp, blueprint_bp, other_bp, custom_bp]



import flask
import flask_webtest
from keg_apps.flaskwtf_bug import BugFixApp, BugApp


class TestViewRouting(object):

    def test_bug_presence(self):
        """ This test will BREAK when the bug is no longer present upstream, at which point this
        fix can be considered for removal.  """
        app = BugApp.testing_prep(FIX_FLASK_WTF_BUG=False)
        client = flask_webtest.TestApp(app)
        client.get('/form')
        assert 'csrf_token' in flask.g

    def test_bug_fix(self):
        app = BugFixApp.testing_prep()
        client = flask_webtest.TestApp(app)
        client.get('/form')
        assert 'csrf_token' not in flask.g




class TestProfile(object):
    # These binds are used to test keg dialect functionality. All other database tests
    # are done on the primary connection, which defaults to SQLite.
    SQLALCHEMY_BINDS = {
        # Change the postgres database connection to a database dedicated to Keg tests.
        'postgres': 'postgresql://postgres:Password12!@localhost/keg_test',
        'mssql': 'mssql+pyodbc_mssql://sa:Password12!@localhost:1433/tempdb'
        '?driver=SQL+Server+Native+Client+11.0',
        'sqlite2': 'sqlite:///'
    }



from flask import current_app  # noqa: F401

from keg.version import VERSION  # noqa: F401
from keg.app import Keg  # noqa: F401
from keg.component import (  # noqa: F401
    KegComponent,
    KegModelComponent,
    KegModelViewComponent,
    KegViewComponent,
)

__all__ = [
    'Keg',
    'KegComponent',
    'KegModelComponent',
    'KegModelViewComponent',
    'KegViewComponent',
    'current_app',
]



from __future__ import absolute_import

from keg.app import Keg
from keg_apps.web.views import all_blueprints


def addkeg(value):
    return '{}Keg'.format(value)


def sayfoo():
    return 'foo'


class WebApp(Keg):
    import_name = 'keg_apps.web'
    use_blueprints = all_blueprints
    keyring_enabled = False

    template_filters = {'addkeg': addkeg}
    template_globals = {'sayfoo': sayfoo}

    def on_init_complete(self):
        # Using .route() in an instance context
        @self.route('/simple3')
        def simple3():
            return 'simple3'


# Using .route() in a class context
@WebApp.route('/simple', endpoint='simple1', methods=['GET', 'POST'])
def simple():
    return 'simple'


@WebApp.route('/simple2', methods=['GET', 'POST'])
def simple2():
    return 'simple2'



from flask.signals import Namespace

_signals = Namespace()

# Call when application initializations is completed.
# Sender will be the application instance.
init_complete = _signals.signal('init-complete')
# Will deprecate app-ready eventually
app_ready = _signals.signal('app-ready')

# Same signals.  Will deprecate config_ready eventually.
config_ready = _signals.signal('config-ready')
config_complete = _signals.signal('config-complete')

# Call when db init_app complete, before any models are visited.
# Note that when this called, app setup is not complete. db is
# the first extension set up, so other extensions, app.visit_modules,
# etc. will not be available.
# Ideal for setting metadata properties.
db_before_import = _signals.signal('db-before-import')

testing_run_start = _signals.signal('testing-run-start')

db_clear_pre = _signals.signal('db-clear-pre')
db_clear_post = _signals.signal('db-clear-post')
db_init_pre = _signals.signal('db-init-pre')
db_init_post = _signals.signal('db-init-post')



from collections import defaultdict
from contextlib import contextmanager
from itertools import chain
import urllib

import click
import flask
import flask.cli

from keg import current_app
from keg.extensions import gettext as _


try:
    import dotenv
except ImportError:
    dotenv = None

try:
    from flask.helpers import get_load_dotenv
except ImportError:
    def get_load_dotenv(default=False):
        return False


class KegAppGroup(flask.cli.AppGroup):
    def __init__(self, create_app, add_default_commands=True, load_dotenv=True, *args, **kwargs):
        self.create_app = create_app
        self.load_dotenv = load_dotenv

        flask.cli.AppGroup.__init__(self, *args, **kwargs)
        if add_default_commands:
            self.add_command(dev_command)

        self._loaded_plugin_commands = False

    def _load_plugin_commands(self):
        if self._loaded_plugin_commands:
            return
        entry_point_iterables = []

        # older Flask
        try:
            import pkg_resources
            entry_point_iterables.extend([
                pkg_resources.iter_entry_points('flask.commands'),
                pkg_resources.iter_entry_points('keg.commands'),
            ])
        except ImportError:
            return

        # newer Flask
        try:
            # uses importlib, but python has some API variations. Let flask handle that.
            from flask.cli import metadata
            entry_point_iterables.extend([
                metadata.entry_points(group="flask.commands"),
                metadata.entry_points(group="keg.commands"),
            ])
        except ImportError:
            pass

        for ep in chain.from_iterable(entry_point_iterables):
            self.add_command(ep.load(), ep.name)
        self._loaded_plugin_commands = True

    def _load_app(self, ctx):
        self._load_plugin_commands()

        info = ctx.ensure_object(flask.cli.ScriptInfo)
        info.load_app()

    def list_commands(self, ctx):
        self._load_app(ctx)
        rv = set(click.Group.list_commands(self, ctx))
        return sorted(rv)

    def get_command(self, ctx, name):
        self._load_app(ctx)
        return click.Group.get_command(self, ctx, name)

    def main(self, *args, **kwargs):
        if get_load_dotenv(self.load_dotenv):
            flask.cli.load_dotenv()

        obj = kwargs.get('obj')
        if obj is None:
            obj = flask.cli.ScriptInfo(create_app=self.create_app)
        kwargs['obj'] = obj
        # TODO: figure out if we want to use this next line.
        #kwargs.setdefault('auto_envvar_prefix', 'FLASK')
        return flask.cli.AppGroup.main(self, *args, **kwargs)


@click.group('develop', help=_('Developer info and utils.'))
def dev_command():
    pass


dev_command.add_command(flask.cli.run_command)
dev_command.add_command(flask.cli.shell_command)


@dev_command.command('routes', short_help=_('List the routes defined for this app.'))
@flask.cli.with_appcontext
def routes_command():
    output = []
    endpoint_len = 0
    methods_len = 0

    # calculate how wide the output columns need to be
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        endpoint_len = max(endpoint_len, len(rule.endpoint))
        methods_len = max(methods_len, len(methods))

    # generate the output
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        line = urllib.parse.unquote("{}   {}   {}".format(
            rule.endpoint.ljust(endpoint_len), methods.ljust(methods_len), rule))
        output.append(line)

    for line in sorted(output):
        click.echo(line)


@dev_command.command('templates', short_help=_('Show paths searched for a template.'))
@flask.cli.with_appcontext
def templates_command():
    jinja_loader = flask.current_app.jinja_env.loader
    paths = defaultdict(list)
    for template in jinja_loader.list_templates():
        for loader, template in jinja_loader._iter_loaders(template):
            for dpath in loader.searchpath:
                paths[dpath].append(template)
    for dpath in sorted(paths.keys()):
        template_paths = sorted(paths[dpath])
        click.echo(dpath)
        for tpath in template_paths:
            click.echo('    {}'.format(tpath))


@dev_command.command('config', short_help=_('List info related to config files, profiles, and'
                     ' values.'))
@flask.cli.with_appcontext
def config_command():
    app = flask.current_app
    config = app.config
    keys = list(config.keys())
    keys.sort()

    click.echo(_('Default config objects:'))
    for val in config.default_config_locations_parsed():
        click.echo('    {}'.format(val))

    click.echo(_('Config file locations:'))
    for val in config.config_file_paths():
        click.echo('    {}'.format(val))

    if config.config_paths_unreadable:
        click.echo(_('Could not access the following config paths:'))
        for path, exc in config.config_paths_unreadable:
            click.echo('    {}: {}'.format(path, str(exc)))

    click.echo(_('Config objects used:'))
    for val in config.configs_found:
        click.echo('    {}'.format(val))

    click.echo(_('Resulting app config (including Flask defaults):'))
    for key in keys:
        click.echo('    {} = {}'.format(key, config[key]))


class DatabaseGroup(click.MultiCommand):

    def list_commands(self, ctx):
        return ['clear', 'init']

    def get_command(self, ctx, name):
        if name == 'init':
            return database_init
        if name == 'clear':
            return database_clear


@dev_command.command('db', cls=DatabaseGroup, invoke_without_command=True,
                     help=_('Lists database related sub-commands.'))
@flask.cli.with_appcontext
@click.pass_context
def database_group(ctx):
    # only take action if no subcommand is involved.
    if ctx.invoked_subcommand is None:
        if not current_app.db_enabled:
            click.echo(_('Database not enabled for this app.  No subcommands available.'))
        else:
            # Database enabled, but no subcommand was given, therefore we want to just show
            # the help message, which would be the default behavior if we had not used the
            # invoke_Without_command option.
            click.echo(ctx.get_help())
            ctx.exit()


@click.command('init', short_help=_('Create all db objects, send related events.'))
@click.option('--clear-first', default=False, is_flag=True,
              help=_('Clear DB of all data and drop all objects before init.'))
@click.option('--yes', default=False, is_flag=True, help="Force confirmation")
@flask.cli.with_appcontext
def database_init(clear_first, yes):
    prompt = clear_first and not yes
    if prompt and not click.confirm(_('Are you sure? You will delete all the data!')):
        click.echo('Database untouched')
        return

    if clear_first:
        current_app.db_manager.db_init_with_clear()
        click.echo(_('Database cleared and initialized'))
    else:
        current_app.db_manager.db_init()
        click.echo(_('Database initialized'))


@click.command('clear', short_help=_('Clear DB of all data and drop all objects.'))
@click.option('--yes', default=False, is_flag=True, help="Force confirmation")
@flask.cli.with_appcontext
def database_clear(yes):
    if yes or click.confirm(_('Are you sure? You will delete all the data!')):
        current_app.db_manager.db_clear()
        click.echo(_('Database cleared'))
    else:
        click.echo(_('Database untouched'))


class CLILoader(object):
    """
        This loader takes care of the complexity of click object setup and instantiation in the
        correct order so that application level CLI options are available before the Keg app is
        instantiated (so the options can be used to configure the app).

        The order of events is:

        - instantiate KegAppGroup
        - KegAppGroup.main() is called
            1. the ScriptInfo object is instantiated
            2. normal click behavior starts, including argument parsing
                - arguments are always parsed due to invoke_without_command=True
            3. during argument parsing, option callbacks are excuted which will result in at least
               the profile name being saved in ScriptInfo().data
            4. Normal click .main() behavior will continue which could include processing commands
               decorated with flask.cli.with_appcontext() or calls to KegAppGroup.list_commands()
               or KegAppGroup.get_command().
                - ScriptInfo.init_app() will be called during any of these operations
                - ScriptInfo.init_app() will call self.create_app() below with the ScriptInfo
                  instance
    """
    def __init__(self, appcls):
        self.appcls = appcls
        # Don't store instance-level vars here.  This object is only instantiated once per Keg
        # sub-class but can be used across multiple app instance creations.  So, anything app
        # instance specific should go on the ScriptInfo instance (see self.options_callback()).

    def create_group(self):
        """ Create the top most click Group instance which is the entry point for any Keg app
            being called in a CLI context.

            The return value of this context gets set on Keg.cli
        """

        return KegAppGroup(
            self.create_app,
            params=self.create_script_options(),
            callback=self.main_callback,
            invoke_without_command=True,
        )

    def create_app(self, script_info=False):
        """ Instantiate our app, sending CLI option values through as needed.

            `script_info:` required for Flask 1.X create_app signature, but not valid in Flask 2.1.
            So keep it as an arg, but don't use it to get the object.  See #163.
        """
        script_info = click.get_current_context().obj
        init_kwargs = self.option_processor(script_info.data)
        return self.appcls().init(**init_kwargs)

    def option_processor(self, cli_options):
        """
            Turn cli_options, which is the result of all parsed options in create_script_options()
            into a dict which will be given as kwargs to the app's .init() method.
        """
        retval = dict(config_profile=cli_options.get('profile', None), config={})
        if cli_options.get('quiet', False):
            retval['config']['KEG_LOG_STREAM_ENABLED'] = False
        return retval

    def create_script_options(self):
        """ Create app level options, ideally that are used to configure the app itself.  """
        return [
            click.Option(['--profile'], is_eager=True, default=None, callback=self.options_callback,
                         help=_('Name of the configuration profile to use.'),),
            click.Option(['--quiet'], is_eager=True, is_flag=True, default=False,
                         callback=self.options_callback,
                         help=_('Set default logging level to logging.WARNING.')),
            click.Option(['--help-all'], is_eager=True, is_flag=True, expose_value=False,
                         callback=self.help_all_callback,
                         help=_('Show all commands with subcommands.')),
        ]

    def options_callback(self, ctx, param, value):
        """ This method is called after argument parsing, after ScriptInfo instantiation but before
            create_app() is called.  It's the only way to get the options into ScriptInfo.data
            before the Keg app instance is instantiated.
        """
        si = ctx.ensure_object(flask.cli.ScriptInfo)
        si.data[param.name] = value

    def help_all_callback(self, ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        formatter = ctx.make_formatter()

        ctx.command.format_usage(ctx, formatter)
        ctx.command.format_help_text(ctx, formatter)
        self.format_options(ctx.command, ctx, formatter)
        self.format_commands(ctx.command, ctx, formatter)
        ctx.command.format_epilog(ctx, formatter)

        click.echo(formatter.getvalue().rstrip('\n'))
        ctx.exit()

    def format_options(self, command, ctx, formatter):
        opts = []
        for param in command.get_params(ctx):
            if param.name == 'help':
                opts.append(('--help', _('Show help message.')))
            elif param.name == 'help_all':
                opts.append(('--help-all', _('Show this message and exit.')))
            else:
                rv = param.get_help_record(ctx)
                if rv is not None:
                    opts.append(rv)

        if opts:
            with formatter.section('Options'):
                formatter.write_dl(opts)

    def format_commands(self, command, ctx, formatter):
        if hasattr(command, 'list_commands'):
            subcommands = command.list_commands(ctx)

            if len(subcommands):
                # allow for 3 times the default spacing
                limit = formatter.width - 6 - max(len(subcommand) for subcommand in subcommands)

                rows = []
                commands = []
                for subcommand in subcommands:
                    cmd = command.get_command(ctx, subcommand)
                    if cmd is None:
                        continue
                    if cmd.hidden:
                        continue
                    commands.append(cmd)
                    help = cmd.get_short_help_str(limit)
                    rows.append((subcommand, help))

                with self.compact_section(formatter, 'Commands'):
                    self.write_dl_with_subcommands(ctx, formatter, rows, commands)

    def write_dl_with_subcommands(self, ctx, formatter, rows, commands, col_max=30, col_spacing=2):
        rows = list(rows)
        widths = click.formatting.measure_table(rows)
        if len(widths) != 2:
            raise TypeError('Expected two columns for definition list')

        first_col = min(widths[0], col_max) + col_spacing

        for (first, second), command in zip(
            click.formatting.iter_rows(rows, len(widths)), commands
        ):
            formatter.write('%*s%s' % (formatter.current_indent, '', first))
            if not second:
                formatter.write('\n')
                continue
            if click.formatting.term_len(first) <= first_col - col_spacing:
                formatter.write(' ' * (first_col - click.formatting.term_len(first)))
            else:
                formatter.write('\n')
                formatter.write(' ' * (first_col + formatter.current_indent))

            text_width = max(formatter.width - first_col - 2, 10)
            lines = iter(click.formatting.wrap_text(second, text_width).splitlines())
            if lines:
                formatter.write(next(lines) + '\n')
                for i, line in enumerate(lines):
                    formatter.write('%*s%s\n' % (
                        first_col + formatter.current_indent, '', line))
            else:
                formatter.write('\n')

            # Add subcommands under current command (recursive)
            with formatter.indentation():
                self.format_commands(command, ctx, formatter)

    @contextmanager
    def compact_section(self, formatter, name):
        formatter.write_heading(name)
        formatter.indent()
        try:
            yield
        finally:
            formatter.dedent()

    def main_callback(self, **kwargs):
        """
            Default Click behavior is to call the help method if no arguments are given to the
            top-most command.  That's good UX, but the way Click impliments it, the help is
            shown before argument parsing takes place.  That's bad, because it means our
            options_callback() is never called and the config profile isn't available when the
            help command calls KegAppGroup.list_commands().

            So, we force Click to always parse the args using `invoke_without_command=True` above,
            but when we do that, Click turns off the automatic display of help.  So, we just
            impliment help-like behavior in this method, which gives us the same net result
            as default Click behavior.
        """
        ctx = click.get_current_context()

        if ctx.invoked_subcommand is not None:
            # A subcommand is present, so arguments were passed.
            return

        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()



import os

import pytest

import keg
from keg.assets import AssetManager, AssetException
from keg_apps.templating.app import TemplatingApp


def setup_module(module):
    TemplatingApp.testing_prep()


class TestAssets(object):

    def test_load_related(self):
        am = AssetManager(keg.current_app)
        am.load_related('assets_in_template.html')

        assert len(am.content['js']) == 1
        asset_name, filepath, content = am.content['js'].pop()
        assert asset_name == 'assets_in_template.js'
        assert filepath.endswith(
            os.path.join('keg_apps', 'templating', 'templates', 'assets_in_template.js')
        )
        assert content.strip() == '//assets_in_template js file'

        assert len(am.content['css']) == 1
        asset_name, filepath, content = am.content['css'].pop()
        assert asset_name == 'assets_in_template.css'
        assert filepath.endswith(
            os.path.join('keg_apps', 'templating', 'templates', 'assets_in_template.css')
        )
        assert content.strip() == '/* assets_in_template css file */'

    def test_load_related_none_found(self):
        with pytest.raises(AssetException) as e:
            am = AssetManager(keg.current_app)
            am.load_related('_not_there.html')
            assert str(e) == 'Could not find related assets for template: _not_there.html'

    def test_load_single_asset(self):
        am = AssetManager(keg.current_app)
        am.load_related('assets_include_single.html')

        assert len(am.content['js']) == 1
        assert len(am.content['css']) == 0



import importlib

import flask


class KegComponent:
    """ Keg components follow the paradigm of flask extensions, and provide some defaults for the
    purpose of setting up model/view structure. Using components, a project may be broken down into
    logical blocks, each having their own entities, blueprints, templates, tests, etc.

    Setup involves:
    - KEG_REGISTERED_COMPONENTS config setting: assumed to be an iterable of importable dotted paths
    - `__component__`: at the top level of each dotted path, this attribute should point to an
    instance of `KegComponent`. E.g. `__component__ = KegComponent('widgets')`

    By default, components will load entities from `db_visit_modules` into metadata and register
    any blueprints specified by `load_blueprints`.

    Blueprints can be created with the helper methods `create_named_blueprint` or `create_blueprint`
    in order to have a configured template folder relative to the blueprint path.

    Use KegModelComponent, KegViewComponent, or KegModelViewComponent for some helpful defaults for
    model/blueprint discovery.

    db_visit_modules: an iterable of dotted paths (e.g. `.mycomponent.entities`,
                      `app.component.extraentities`) where Keg can find the entities for this
                      component to load them into the metadata.

    .. note:: Normally this is not explicitly required but can be useful in cases where imports
       won't reach that file.

    .. note:: This can accept relative dotted paths (starts with `.`) and it will prepend the
       component python package determined by Keg when instantiating the component. You can also
       pass absolute dotted paths and no alterations will be performed.

    load_blueprints: an iterable of tuples, each having a dotted path (e.g. `.mycomponent.views`,
                     `app.component.extraviews`) and the blueprint attribute name to load and
                     register on the app. E.g. `(('.views', 'component_bp'), )`

    .. note:: This can accept relative dotted paths (starts with `.`) and it will prepend the
       component python package determined by Keg when instantiating the component. You can also
       pass absolute dotted paths and no alterations will be performed.

    template_folder: string to be passed for template config to blueprints created via the component
    """
    db_visit_modules = tuple()
    load_blueprints = tuple()
    template_folder = 'templates'

    def __init__(self, name, app=None, db_visit_modules=None, load_blueprints=None,
                 template_folder=None, parent_path=None):
        self.name = name
        # Allow customization of the defaults in the constructor
        self.db_visit_modules = db_visit_modules or self.db_visit_modules
        self.load_blueprints = load_blueprints or self.load_blueprints
        self.template_folder = template_folder or self.template_folder
        if app:
            # Not really intended to be used this way, but it fits the flask extension paradigm
            # and could conceivably be set up in lieu of KEG_REGISTERED_COMPONENTS
            self.init_app(app, parent_path=parent_path)

    def init_app(self, app, parent_path=None):
        # parent_path gets used as an absolute parent for the relative import paths of
        # model/blueprints.
        # E.g. if relative_dotted_path is `my_app.components.widget` and one of the relative import
        # paths is `.model.entities`, the full import is `my_app.components.widget.model.entities`
        self.init_config(app)
        self.init_db(parent_path)
        self.init_blueprints(app, parent_path)

    def init_config(self, app):
        # Components may define their own config defaults
        pass

    def init_db(self, parent_path):
        # Intent is to import the listed modules, so their entities are registered in metadata
        for dotted_path in self.db_visit_modules:
            import_name = dotted_path
            if import_name.startswith('.'):
                import_name = f'{parent_path}{dotted_path}'
            importlib.import_module(import_name)

    def init_blueprints(self, app, parent_path):
        # Register any blueprints that are listed by path in load_blueprints
        for bp_path, bp_attr in self.load_blueprints:
            import_name = bp_path
            if import_name.startswith('.'):
                import_name = f'{parent_path}{bp_path}'
            mod_imported = importlib.import_module(import_name)
            app.register_blueprint(getattr(mod_imported, bp_attr))

    def create_blueprint(self, *args, **kwargs):
        """Make a flask blueprint having a template folder configured.

        Generally, args and kwargs provided will be passed to the blueprint constructor, with
        the following exceptions:

        - template_folder kwarg defaults to the component's template_folder if not provided
        - blueprint_cls kwarg may be used to specify an alternative to flask.Blueprint
        """
        kwargs['template_folder'] = kwargs.get('template_folder', self.template_folder)
        blueprint_cls = kwargs.pop('blueprint_cls', flask.Blueprint)
        bp = blueprint_cls(*args, **kwargs)

        return bp

    def create_named_blueprint(self, *args, **kwargs):
        # Make a flask blueprint named with the component name, having a template folder configured
        return self.create_blueprint(self.name, *args, **kwargs)


class ModelMixin:
    db_visit_modules = ('.model.entities', )


class ViewMixin:
    load_blueprints = (('.views', 'component_bp'), )


class KegModelComponent(ModelMixin, KegComponent):
    pass


class KegViewComponent(ViewMixin, KegComponent):
    pass


class KegModelViewComponent(ModelMixin, ViewMixin, KegComponent):
    pass



from __future__ import absolute_import

import sys


def fix_sys_path():
    """
        When running our apps with python like:

            python keg_apps/cli.py

        `keg_apps/` gets put on the front of sys.path.  So, when other packages, like blazeutils
        call `import logging` keg_apps.logging is imported.

        Fix this situation by simply removing keg_apps from sys.path.  It doesn't need to be there
        for anything we are doing.
    """
    if sys.path[0].endswith('keg_apps'):
        sys.path.pop(0)



import tempfile
from unittest import mock

import flask
import pytest

from keg.testing import inrequest
from keg.utils import pymodule_fpaths_to_objects
from keg_apps.web.app import WebApp


class TestUtils(object):

    def test_pymodule_fpaths_to_objects(self):
        with tempfile.NamedTemporaryFile(delete=False) as fh:
            fpath1 = fh.name
            fh.write(b'foo1="bar"')

        with tempfile.NamedTemporaryFile(delete=False) as fh:
            fpath2 = fh.name
            fh.write(b'foo2="bar"')

        with tempfile.NamedTemporaryFile() as fh:
            fpath3 = fh.name

        result = pymodule_fpaths_to_objects([fpath1, fpath2, fpath3])

        fpath, fpath_objs, exc = result.pop(0)
        assert fpath == fpath1
        assert fpath_objs['foo1'] == 'bar'

        fpath, fpath_objs, exc = result.pop(0)
        assert fpath == fpath2
        assert fpath_objs['foo2'] == 'bar'

        # result should now have fpath3. Error case is tested separately.
        assert len(result) == 1

    @pytest.mark.parametrize('error', (FileNotFoundError, IsADirectoryError, PermissionError))
    @mock.patch('keg.utils.open')
    def test_pymodule_fpaths_to_objects_error(self, m_open, error):
        m_open.side_effect = error

        result = pymodule_fpaths_to_objects(['some-path.py'])
        fpath, fpath_objs, exc = result.pop(0)
        assert fpath == 'some-path.py'
        assert fpath_objs is None
        assert isinstance(exc, error)


class TestInRequest:
    @classmethod
    def setup_class(cls):
        WebApp.testing_prep()

    @inrequest('/mypath?foo=bar&baz=boo')
    def test_in_request_args(self):
        assert flask.request.args['foo'] == 'bar'

    @inrequest('/mypath?foo=bar&baz=boo')
    def test_request_args_immutable(self):
        with pytest.raises(TypeError, match='immutable'):
            flask.request.args['foo'] = 'moo'

    def test_request_args_mutated(self):
        def args_modifier(args_dict):
            args_dict['baz'] = 'custom-value'

        with inrequest('/mypath?foo=bar&baz=boo', args_modifier=args_modifier):
            assert flask.request.args['foo'] == 'bar'
            assert flask.request.args['baz'] == 'custom-value'

            # ensure args return to immutable after patch
            with pytest.raises(TypeError, match='immutable'):
                flask.request.args['foo'] = 'moo'



from __future__ import absolute_import
from __future__ import print_function

from keg.app import Keg


class DB2App(Keg):
    import_name = __name__
    # Silence the keyring warning messages
    keyring_enabled = False
    db_enabled = True
    db_visit_modules = []



from keg.testing import WebBase

from keg_apps.web.app import WebApp


class TestViewTemplating(WebBase):
    appcls = WebApp

    def test_templating(self):
        resp = self.testapp.get('/template1')
        assert resp.text == 'Hello world!'

    def test_custom_name(self):
        resp = self.testapp.get('/template2')
        assert resp.text == 'template-too'


class TestJinjaInit(WebBase):
    appcls = WebApp

    def test_jinja_filter(self):
        resp = self.testapp.get('/jinja')
        assert resp.text == 'hello Keg foo'



from flask.ctx import RequestContext
from keg.assets import AssetManager


class KegRequestContext(RequestContext):

    def __init__(self, app, environ, request=None):
        super(KegRequestContext, self).__init__(app, environ, request)
        self.assets = AssetManager(app)



from __future__ import absolute_import

import flask
from flask_wtf import FlaskForm
from keg.app import Keg
from wtforms import StringField


class BugApp(Keg):
    import_name = __name__
    keyring_enabled = False


# Has to be a second app b/c the fix depends on a configuration value passed through testing_prep()
# which will only apply once per App class due to how testing.ContextManager works.
class BugFixApp(Keg):
    import_name = __name__
    keyring_enabled = False


class MyForm(FlaskForm):
    name = StringField('name')


@BugApp.route('/form')
@BugFixApp.route('/form')
def form():
    form = MyForm()
    form_tpl = '''
    {{ form.csrf_token }}
    {{ form.name(size=20) }}
    '''
    return flask.render_template_string(form_tpl, form=form)



from blazeutils.strings import randchars

from keg.db import db


class Blog(db.Model):
    __tablename__ = 'blogs'

    id = db.Column(db.Integer, primary_key=True)
    title = db.Column(db.Unicode(50), unique=True, nullable=False)

    @classmethod
    def testing_create(cls):
        blog = Blog(title=randchars())
        db.session.add(blog)
        db.session.commit()
        return blog



from io import StringIO
import logging
from unittest import mock

from keg import signals
from keg_apps.logging import LoggingApp, log


class TestLogging(object):

    def test_stream_handler(self):
        with mock.patch('sys.stderr', new_callable=StringIO) as m_stderr:
            LoggingApp().init(use_test_profile=True)
            log.warning('test warn log')
            log.info('test info log')

        output = m_stderr.getvalue().strip().splitlines()
        assert len(output) == 2, output

        assert output[0] == 'WARNING - keg_apps.logging - test warn log'
        assert output[1] == 'INFO - keg_apps.logging - test info log'

    def test_disable_handlers(self):
        app = LoggingApp()

        # It's important to use connect_via() here or the signal will apply to all tests, which
        # produces intermittent test failure, which is hard to troubleshoot, which is bad.
        @signals.config_ready.connect_via(app)
        def apply_config(app):
            app.config['KEG_LOG_STREAM_ENABLED'] = False
            app.config['KEG_LOG_SYSLOG_ENABLED'] = False

        with mock.patch('keg.logging.Logging.init_syslog') as m_init_syslog, \
                mock.patch('keg.logging.Logging.init_stream') as m_init_stream:
            app.init(use_test_profile=True)
            m_init_syslog.assert_not_called()
            m_init_stream.assert_not_called()

    def test_syslog_handler(self):
        with mock.patch('keg.logging.SysLogHandler.emit') as m_emit:
            LoggingApp().init(use_test_profile=True, config={'KEG_LOG_SYSLOG_ENABLED': True})
            log.warning('test warn log')
            log.info('test info log')

        calls = m_emit.call_args_list
        assert len(calls) == 2

        args, kwargs = calls[0]
        log_record = args[0]
        assert log_record.message == 'test warn log'

        args, kwargs = calls[1]
        log_record = args[0]
        assert log_record.message == 'test info log'


class TestJsonFormatter(object):
    def setup_method(self):
        self.logger = logging.getLogger('logging-test')
        self.logger.setLevel(logging.DEBUG)
        self.buffer = StringIO()

        self.logHandler = logging.StreamHandler(self.buffer)
        self.logger.addHandler(self.logHandler)

    def test_default_prefix(self):
        app = LoggingApp().init(use_test_profile=True)
        jf = app.logging.create_json_formatter()
        self.logHandler.setFormatter(jf)

        self.logger.info("foo")

        output = self.buffer.getvalue()
        assert output.startswith('@cee:')

    def test_config_prefix_overrides_default(self):
        config = {'KEG_LOG_JSON_FORMATTER_KWARGS': {'prefix': '@bar'}}
        app = LoggingApp().init(use_test_profile=True, config=config)
        jf = app.logging.create_json_formatter()
        self.logHandler.setFormatter(jf)

        self.logger.info("foo")

        output = self.buffer.getvalue()
        assert output.startswith('@bar')



import flask_sqlalchemy as fsa
import sqlalchemy as sa
import sqlalchemy.event as sa_event

from keg.signals import (
    db_before_import,
    db_clear_post,
    db_clear_pre,
    db_init_post,
    db_init_pre,
    testing_run_start,
)
from keg.utils import visit_modules


class KegSQLAlchemy(fsa.SQLAlchemy):
    def _apply_driver_defaults(self, options, app):
        """Override some driver specific settings"""
        super_return_value = None
        if hasattr(super(), '_apply_driver_defaults'):
            super_return_value = super()._apply_driver_defaults(options, app)

        # Turn on SA pessimistic disconnect handling by default:
        # http://docs.sqlalchemy.org/en/latest/core/pooling.html#disconnect-handling-pessimistic
        options.setdefault('pool_pre_ping', True)

        # While this isn't an engine options change, it is in the domain of db engine
        # setup and so belongs here.
        if app.config.get('KEG_SQLITE_ENABLE_FOREIGN_KEYS'):
            sa_event.listens_for(sa.engine.Engine, 'connect')(self._set_sqlite_pragma)

        return super_return_value

    def apply_driver_hacks(self, app, info, options):
        """This method is renamed to _apply_driver_defaults in flask-sqlalchemy 3.0"""
        super_return_value = super().apply_driver_hacks(app, info, options)

        # follow the logic to set some defaults, but the super won't exist there
        self._apply_driver_defaults(options, app)

        return super_return_value

    def get_engine(self, app=None, bind=None):
        if not hasattr(self, '_app_engines'):
            # older version of flask-sqlalchemy, we can just call super
            return super().get_engine(app=app, bind=bind)

        # More recent flask-sqlalchemy, use the cached engines directly.
        # Note: we don't necessarily have an app context active here, depending
        # on if this is being called during app init. But if we attempt to access
        # the underlying cache directly, we get a weak ref error.
        with app.app_context():
            return self.engines[bind]

    def get_engines(self, app):
        # the default engine doesn't have a bind
        retval = [(None, self.get_engine(app))]

        bind_names = app.config['SQLALCHEMY_BINDS']
        # The default value of SQLALCHEMY_BINDS is None and the key is present b/c of
        # Flask-SQLAlchemy defaults.  So, only process the binds if they are not None.
        if bind_names is not None:
            for bind_name in bind_names:
                retval.append((bind_name, self.get_engine(app, bind=bind_name)))

        return retval

    def _set_sqlite_pragma(self, connection, conn_record):
        # Sets a pragma to tell sqlite to not ignore FK constraints.
        from sqlite3 import Connection as SQLite3Connection
        if isinstance(connection, SQLite3Connection):
            cursor = connection.cursor()
            cursor.execute("PRAGMA foreign_keys=ON;")
            cursor.close()


db = KegSQLAlchemy()

# put this import after the above db assignment to avoid circular reference issues
from .dialect_ops import DialectOperations  # noqa


class DatabaseManager(object):
    """
        A per-app-instance utility class that managers all common operations for a Keg app.

        Binds & Dialects
        ----------------

        Flask SQLAlchemy handles multiple DB connections per application through the use of "binds."
        When an application wants to communicate events or initiate activites, this manager will
        will handle distributing those events and activities to all database connections bound
        to the application.

        Furthermore, this manager delegates to DialectOperations instances to run the events and
        activities in ways that are specific to the type of RDBMS being used (when needed).
    """

    def __init__(self, app):
        self.app = app
        self.dialect_opts = app.config['KEG_DB_DIALECT_OPTIONS']

        self.init_app()
        self.init_events()

    def init_app(self):
        db.init_app(self.app)
        db_before_import.send(self.app)
        visit_modules(self.app.db_visit_modules, self.app.import_name)

    def init_events(self):
        testing_run_start.connect(self.on_testing_start, sender=self.app)

    def bind_dialect(self, bind_name):
        engine = db.get_engine(self.app, bind=bind_name)
        return DialectOperations.create_for(engine, bind_name, self.dialect_opts)

    def all_bind_dialects(self):
        """
            For each database connection (bind) in this application, yield a DialectOperations
            instance corresponding to the type of RDBMS the bind is connecting to.
        """
        for bind_name, engine in db.get_engines(self.app):
            yield DialectOperations.create_for(engine, bind_name, self.dialect_opts)

    def on_testing_start(self, app):
        self.db_init_with_clear()

    def create_all(self):
        for dialect in self.all_bind_dialects():
            dialect.create_all()

    def drop_all(self):
        db.session.remove()
        for dialect in self.all_bind_dialects():
            dialect.drop_all()

    # The methods that follow will trigger application events.
    def db_init_with_clear(self):
        self.db_clear()
        self.db_init()

    def db_init(self):
        db_init_pre.send(self.app)
        self.create_all()
        db_init_post.send(self.app)

    def db_clear(self):
        db_clear_pre.send(self.app)
        self.drop_all()
        db_clear_post.send(self.app)



from blazeutils.strings import randchars

from keg.db import db


####
# The next several entities exist to facilitate testing of the way Keg handles multiple database
# binds and dialects.
####

class PGDud(db.Model):
    __tablename__ = 'pgdud'
    __bind_key__ = 'postgres'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50), unique=True, nullable=False)


class PGDud2(db.Model):
    __tablename__ = 'pgdud'
    __bind_key__ = 'postgres'
    __table_args__ = {'schema': 'fooschema'}

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50), unique=True, nullable=False)

    @classmethod
    def testing_create(cls):
        dud = cls(name=randchars())
        db.session.add(dud)
        db.session.commit()
        return dud


class SADud(db.Model):
    __tablename__ = 'sadud'
    __bind_key__ = 'sqlite2'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50), unique=True, nullable=False)


class MSSQLDud(db.Model):
    __tablename__ = 'mssqldud'
    __bind_key__ = 'mssql'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50), unique=True, nullable=False)


class MSSQLDud2(db.Model):
    __tablename__ = 'mssqldud'
    __bind_key__ = 'mssql'
    __table_args__ = {'schema': 'fooschema'}

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50), unique=True, nullable=False)

    @classmethod
    def testing_create(cls):
        dud = cls(name=randchars())
        db.session.add(dud)
        db.session.commit()
        return dud



MORPHI_PACKAGE_NAME = 'keg_apps'

# begin morphi boilerplate
try:
    import morphi
except ImportError:
    morphi = None

if morphi:
    from morphi.messages import Manager
    from morphi.registry import default_registry

    translation_manager = Manager(package_name=MORPHI_PACKAGE_NAME)
    default_registry.subscribe(translation_manager)

    gettext = translation_manager.gettext
    lazy_gettext = translation_manager.lazy_gettext
    lazy_ngettext = translation_manager.lazy_ngettext
    ngettext = translation_manager.ngettext

else:
    translation_manager = None

    def gettext(message, **variables):
        if variables:
            return message.format(**variables)

        return message

    def ngettext(singular, plural, num, **variables):
        variables.setdefault('num', num)

        if num == 1:
            return gettext(singular, **variables)

        return gettext(plural, **variables)

    lazy_gettext = gettext
    lazy_ngettext = ngettext



import os.path as osp
from setuptools import setup, find_packages

cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'readme.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

version_fpath = osp.join(cdir, 'keg', 'version.py')
version_globals = {}
with open(version_fpath) as fo:
    exec(fo.read(), version_globals)

setup(
    name="Keg",
    version=version_globals['VERSION'],
    description=("A web framework built on Flask & SQLAlchemy."
                 "  Somewhere North of Flask but South of Django."),
    long_description='\n\n'.join((README, CHANGELOG)),
    author="Randy Syring",
    author_email="randy.syring@level12.io",
    url='https://github.com/level12/keg',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    license='BSD',
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'appdirs',
        'BlazeUtils',
        'blinker',
        'Click>=3.0',
        'Flask',
        'Flask-SQLAlchemy',
        'python-json-logger',
        'sqlalchemy',
    ],
    extras_require={
        'tests': [
            'flask-webtest',
            'flask-wtf',
            'sqlalchemy_pyodbc_mssql',
            'pyodbc==4.0.34',
            'pytest',
            'pytest-cov',
            'python-dotenv',
            'psycopg2-binary',
            'mock',
        ],
        'i18n': [
            'morphi>=0.2.0'
        ],
    }
)



import os
from unittest import mock

from keg.app import Keg
from keg.config import Config
from keg.testing import cleanup_app_contexts, invoke_command
from keg_apps.profile.cli import ProfileApp


class TestConfigDefaults(object):

    def test_default_profile(self):
        app = Keg(__name__).init()
        assert app.config.profile is None

        # this key comes from Keg's DefaultProfile
        assert 'KEG_ENDPOINTS' in app.config

    def test_init_configs(self):
        app = Keg(__name__, config={'ABC': 123}).init()
        assert app.config['ABC'] == 123

        app = Keg(__name__).init(config={'ABC': 123})
        assert app.config['ABC'] == 123

        class TestingPrepConfigApp(Keg):
            import_name = __name__
        app = TestingPrepConfigApp.testing_prep(ABC=123)
        assert app.config['ABC'] == 123

    def test_app_params_profile(self):
        app = Keg(__name__).init(config_profile='Foo', use_test_profile=True)
        assert app.config.profile == 'Foo'

    @mock.patch.dict('os.environ', {'TEMPAPP_CONFIG_PROFILE': 'bar'})
    def test_environ_profile(self):
        app = Keg('tempapp').init(use_test_profile=True)
        assert app.config.profile == 'bar'

    def test_profile_from_config_file(self):
        config_file_objs = [('/fake/path', dict(DEFAULT_PROFILE='baz'))]
        config = Config('', {})
        config.init_app(None, __name__, '', False, config_file_objs)
        assert config.profile == 'baz'

    def test_profile_from_fpaths(self):
        fpath_results = [
            ('foo.py', dict(DEFAULT_PROFILE='baz'), None),
            ('boo.py', None, FileNotFoundError()),
        ]
        config = Config('', {})
        with mock.patch(
            'keg.config.pymodule_fpaths_to_objects', autospec=True, spec_set=True
        ) as m_fpath:
            m_fpath.return_value = fpath_results
            config.init_app(None, __name__, '', False)
        assert config.profile == 'baz'
        assert config.config_file_objs == [('foo.py', dict(DEFAULT_PROFILE='baz'))]
        assert config.config_paths_unreadable[0][0] == 'boo.py'
        assert isinstance(config.config_paths_unreadable[0][1], FileNotFoundError)

    def test_default_config_objects_no_profile(self):
        config = Config('', {})
        config.init_app(None, 'fakeapp', '', False)

        # No profile given, so only the default profiles from Keg and the App should be tried.
        expected = [
            'keg.config.DefaultProfile',
            'fakeapp.config.DefaultProfile'
        ]
        assert config.default_config_locations_parsed() == expected

    def test_default_config_objects_with_profile(self):
        config = Config('', {})

        config.init_app('SomeProfile', 'fakeapp', '', False)

        expected = [
            'keg.config.DefaultProfile',
            'keg.config.SomeProfile',
            'fakeapp.config.DefaultProfile',
            'fakeapp.config.SomeProfile',
        ]
        assert config.default_config_locations_parsed() == expected

    def test_config_substitutions(self):
        config = Config('', {})
        config['testvalue'] = '{not there}'
        config.init_app(None, 'fakeapp', '', False)
        assert config['testvalue'] == '{not there}'


class TestProfileLoading(object):

    def test_app_init(self):
        """ Even with an environment set, the explict value passed into init() should override. """
        # the direct setting should override the environment
        kwargs = {'KEG_APPS.PROFILE_CONFIG_PROFILE': 'EnvironmentProfile'}
        with mock.patch.dict(os.environ, **kwargs):
            app = ProfileApp().init(config_profile='AppInitProfile', use_test_profile=True)
        assert app.config['PROFILE_FROM'] == 'app-init'

    def test_config_file_default(self):
        """
            No explicit profile and no environment value results in the default coming from
            configuration profiles.
        """
        app = ProfileApp().init()
        assert app.config['PROFILE_FROM'] == 'config-file-default'

    def test_testing_default(self):
        """
            Make sure the testing profile is used when needed.
        """
        app = ProfileApp().init(use_test_profile=True)
        assert app.config['PROFILE_FROM'] == 'testing-default'

        app = ProfileApp.testing_prep()
        assert app.config['PROFILE_FROM'] == 'testing-default'

    def test_environment(self):
        """
            The environment override should work even when in a testing context.
        """
        kwargs = {'KEG_APPS_PROFILE_CONFIG_PROFILE': 'EnvironmentProfile'}
        with mock.patch.dict(os.environ, **kwargs):
            app = ProfileApp().init()
            assert app.config['PROFILE_FROM'] == 'environment'

            app = ProfileApp().init(use_test_profile=True)
            assert app.config['PROFILE_FROM'] == 'environment'

    def test_invoke_command_for_testing(self):
        """
            Using testing.invoke_command() should use a testing profile by default.
        """
        cleanup_app_contexts()
        resp = invoke_command(ProfileApp, 'show-profile')
        assert 'testing-default' in resp.output

    def test_invoke_command_with_environment(self):
        """
            Environement overrides should still take priority for invoke_command() usage.
        """
        cleanup_app_contexts()
        resp = invoke_command(ProfileApp, 'show-profile',
                              env={'KEG_APPS_PROFILE_CONFIG_PROFILE': 'EnvironmentProfile'})
        assert 'environment' in resp.output




class TestProfile(object):
    KEG_DB_DIALECT_OPTIONS = {
        'postgresql.schemas': ('public', 'fooschema'),
        'mssql.schemas': ('fooschema', ),
    }

    # These binds are used to test keg dialect functionality. All other database tests
    # are done on the primary connection, which defaults to SQLite.
    SQLALCHEMY_BINDS = {
        # Change the postgres database connection to a database dedicated to Keg tests.
        'postgres': 'postgresql://user:password@localhost/database',
        'sqlite2': 'sqlite:///'
    }



from __future__ import absolute_import

from keg.app import Keg


class CLI2App(Keg):
    import_name = 'keg_apps.cli2'

    def on_init_complete(self):
        # Want to test when the CLI command is registered after the app is initialized
        import keg_apps.cli2.cli  # noqa


if __name__ == '__main__':
    # This import prevents a nasty bug where there are actually two classes __main__.CLI2App
    # and the one with a full module path.
    import keg_apps.cli2.app as app
    app.CLI2App.cli.main()




class DefaultProfile(object):
    KEG_KEYRING_ENABLE = False

    KEG_REGISTERED_COMPONENTS = {
        'keg_apps.db.blog',
    }


class TestProfile(object):
    KEG_DB_DIALECT_OPTIONS = {
        'postgresql.schemas': ('public', 'barschema'),
        'bind.postgres.schemas': ('public', 'fooschema'),
        'mssql.schemas': ('fooschema', ),
    }

    # These binds are used to test keg dialect functionality in CircleCI. All other database tests
    # are done on the primary connection, which defaults to SQLite.
    #
    # Appveyor: see user-config-tpl-appveyor.py
    # Developers: see instructions in readme for setup of local config file
    SQLALCHEMY_BINDS = {
        # This string will connect to the postgresql DB setup in docker-compose.yaml and
        # work in CircleCI.
        'postgres': 'postgresql://postgres@127.0.0.1:54321/postgres',
        'mssql': 'mssql+pyodbc_mssql://sa:Password12!@127.0.0.1:14331/tempdb'
        '?driver=ODBC+Driver+17+for+SQL+Server',
        'sqlite2': 'sqlite:///'
    }



from __future__ import absolute_import
import flask

from keg.web import BaseView

from keg_apps.extensions import lazy_gettext as _


blueprint = flask.Blueprint('templating', __name__)


class BaseView(BaseView):
    blueprint = blueprint


class Template1(BaseView):
    def get(self):
        self.assign('name', _('world'))


class Template2(BaseView):
    template_name = 'templating/template-two.html'

    def get(self):
        pass


class Jinja(BaseView):

    def get(self):
        pass



import pytest
from unittest import mock

from keg import current_app
from keg.db import db

from keg_apps.db.app import DBApp


def setup_module(module):
    DBApp.testing_prep()


@pytest.fixture(autouse=True)
def db_session_prep():
    """
        Rollback the session after every test.
    """
    db.session.rollback()


class DialectExam(object):
    bind_name = None
    obj_names_sql = None

    @classmethod
    def setup_class(cls):
        # passing None as the app just because on_testing_start() doesn't use it
        current_app.db_manager.on_testing_start(None)

    @classmethod
    def dialect(cls):
        return current_app.db_manager.bind_dialect(cls.bind_name)

    def engine(self):
        return db.get_engine(current_app, self.bind_name)

    def obj_names(self):
        records = self.engine().execute(self.obj_names_sql).fetchall()
        return set([record[0] for record in records])

    def create_objs(self):
        self.engine().execute('create table foo(label varchar(20))')
        self.engine().execute('create view vfoo as select * from foo')
        return 2

    def test_database_clearing(self):
        obj_count = self.create_objs()
        # Expected count after creation is the number of objects created specifically for this test
        # plus the number of tables created for each entity defined in this test app.
        assert len(self.obj_names()) == obj_count + self.entity_count
        self.dialect().drop_all()
        assert len(self.obj_names()) == 0

    def test_create_all(self):
        self.dialect().drop_all()
        self.dialect().create_all()
        assert len(self.obj_names()) == self.entity_count


class TestSQLite(DialectExam):
    bind_name = 'sqlite2'
    obj_names_sql = "select name from sqlite_master where type in ('table', 'view')"
    # How many entities exist for this bind in the application?
    entity_count = 1


class TestPostgreSQL(DialectExam):
    bind_name = 'postgres'
    obj_names_sql = ("SELECT table_schema || '.' || table_name FROM INFORMATION_SCHEMA.tables"
                     " WHERE table_type in ('BASE TABLE', 'VIEW')"
                     " AND table_schema in ('public', 'fooschema')")
    # How many entities exist for this bind in the application?
    entity_count = 2

    def create_objs(self):
        self.engine().execute('create table foo(label varchar(20))')
        self.engine().execute('create view vfoo as select * from foo')
        self.engine().execute('create table fooschema.bar(label varchar(20))')
        self.engine().execute('create view fooschema.vbar as select * from fooschema.bar')
        return 4

    def test_create_all_without_prepping(self):
        """Test that create_all creates necessary schemas"""
        self.dialect().drop_all()
        self.dialect().create_all()
        assert len(self.obj_names()) == self.entity_count

    def test_create_all_existing_schemas(self):
        """Test that create_all can be called when schemas already exist"""
        self.dialect().drop_all()
        self.dialect().create_all()
        self.dialect().create_all()
        assert len(self.obj_names()) == self.entity_count

    def test_options_follow_bind_dialect_specificity(self):
        """Test that create_all creates necessary schemas from the dialect option"""
        dialect_config = current_app.config.get('KEG_DB_DIALECT_OPTIONS').copy()
        del dialect_config['bind.postgres.schemas']
        with mock.patch.object(current_app.db_manager, 'dialect_opts', dialect_config):
            dialect = self.dialect()
            assert dialect.opt_schemas == ('public', 'barschema')
        dialect = self.dialect()
        assert dialect.opt_schemas == ('public', 'fooschema')


class TestMicrosoftSQL(DialectExam):
    bind_name = 'mssql'
    # MSSQL supports the INFORMATION_SCHEMA standard
    obj_names_sql = ("SELECT table_schema + '.' + table_name FROM INFORMATION_SCHEMA.tables"
                     " WHERE table_type in ('BASE TABLE', 'VIEW')"
                     " AND table_schema in ('dbo', 'fooschema')")
    # How many entities exist for this bind in the application?
    entity_count = 2

    def setup_method(self, method):
        if 'mssql' not in current_app.config.get('SQLALCHEMY_BINDS', {}):
            pytest.skip('cannot test missing bind for mssql')

    def create_objs(self):
        self.engine().execute('create table foo(label varchar(20))')
        self.engine().execute('create view vfoo as select * from foo')
        self.engine().execute('create table fooschema.bar(label varchar(20))')
        self.engine().execute('create view fooschema.vbar as select * from fooschema.bar')
        return 4

    def test_create_all_without_prepping(self):
        """Test that create_all creates necessary schemas"""
        self.dialect().drop_all()
        self.dialect().create_all()
        assert len(self.obj_names()) == self.entity_count

    def test_create_all_existing_schemas(self):
        """Test that create_all can be called when schemas already exist"""
        self.dialect().drop_all()
        self.dialect().create_all()
        self.dialect().create_all()
        assert len(self.obj_names()) == self.entity_count



from collections import defaultdict
import pathlib

from jinja2 import TemplateNotFound

from keg.extensions import lazy_gettext as _


class AssetException(Exception):
    pass


class AssetManager(object):
    """
        A per-request helper object for managing assets related to a request/response cycle.
    """

    def __init__(self, app):
        self.app = app
        self.env = self.app.jinja_env
        self.content = defaultdict(list)

    def load_asset(self, asset_name):
        asset_type = pathlib.Path(asset_name).suffix.lstrip('.')
        try:
            contents, filename, _ = self.env.loader.get_source(self.env, asset_name)
            self.content[asset_type].append((asset_name, filename, contents))
            return True
        except TemplateNotFound:
            return False

    def load_related(self, template_name):
        js_asset_name = str(pathlib.PurePosixPath(template_name).with_suffix('.js'))
        js_found = self.load_asset(js_asset_name)
        css_asset_name = str(pathlib.PurePosixPath(template_name).with_suffix('.css'))
        css_found = self.load_asset(css_asset_name)
        if not js_found and not css_found:
            raise AssetException(_('Could not find related assets for template: {template_name}',
                                   template_name=template_name))

    def combine_content(self, asset_type):
        output = []
        for asset_name, filename, contents in self.content[asset_type]:
            comment = '/********************* asset: {} *********************/'.format(asset_name)
            output.append(comment)
            output.append(contents)
        return '\n\n'.join(output)
